import tensorflow as tf
from tensorflow import keras
from teras.losses.saint import info_nce_loss, denoising_loss


@keras.saving.register_keras_serializable(package="teras.layerflow.models.backbones")
class SAINT(keras.Model):
    """
    SAINT architecture with LayerFlow design.
    It proposed by Gowthami Somepalli et al.
    in the paper,
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.

    SAINT performs attention over both rows and columns.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset.

        encoder: ``keras.layers.Layer``,
            An instance of `SAINTEncoder` layer to encode the features embeddings,
            or any layer that can work in its place for that purpose.
            If None, a ``SAINTEncoder`` layer with default values will be used.
            You can import the ``SAINTEncoder`` layer as follows,
                >>> from teras.layerflow.layers import SAINTEncoder

        categorical_feature_embedding: ``keras.layers.Layer``,
            An instance of ``CategoricalFeatureEmbedding`` layer to embedd categorical features
            or any layer that can work in its palce for that purpose.
            If None, a ``CategoricalFeatureEmbedding`` layer with default values will be used.
            You can import the ``CategoricalFeatureEmbedding`` layer as follows,
                >>> from teras.layers import CategoricalFeatureEmbedding

        numerical_feature_embedding: ``keras.layers.Layer``,
            An instance of ``SAINTNumericalFeatureEmbedding`` layer to embedd numerical features
            or any layer that can work in its place for that purpose.
            If None, a ``SAINTNumericalFeatureEmbedding`` layer with default values will be used.
            You can import the ``SAINTNumericalFeatureEmbedding`` layer as follows,
                >>> from teras.layers import SAINTNumericalFeatureEmbedding

        head: ``keras.layers.Layer``,
            An instance of either ``ClassificationHead`` or ``RegressionHead`` layers,
            depending on the task at hand.

            REMEMBER: In case you're using this model as a base model for pretraining, you MUST leave
            this argument as None.

            You can import the ``ClassificationHead`` and ``RegressionHead`` layers as follows,
                >>> from teras.layers import ClassificationHead
                >>> from teras.layers import RegressionHead
    """
    def __init__(self,
                 input_dim: int,
                 categorical_feature_embedding: keras.layers.Layer = None,
                 numerical_feature_embedding: keras.layers.Layer = None,
                 encoder: keras.layers.Layer = None,
                 head: keras.layers.Layer = None,
                 **kwargs):
        if categorical_feature_embedding is None and numerical_feature_embedding is None:
            raise ValueError("Both `categorical_feature_embedding` and `numerical_feature_embedding` "
                             "cannot be None at the same time as a tabular dataset must contains "
                             "features of at least one of the types if not both. ")

        if encoder is None:
            raise ValueError("`encoder` cannot be None. Please pass an instance of `SAINTEncoder` layer. "
                             "You can import it as, `from teras.layerflow.layers import SAINTEncoder``")

        inputs = keras.layers.Input(shape=(input_dim,))

        if categorical_feature_embedding is None:
            # then there's only numerical features because we already
            # have a check above that both categorical and numerical embeddings
            # cannot be None at the same time
            feature_embeddings = numerical_feature_embedding(inputs)
        else:
            categorical_embeddings = categorical_feature_embedding(inputs)
            numerical_embeddings = numerical_feature_embedding(inputs)
            feature_embeddings = keras.layers.Concatenate(axis=1)([categorical_embeddings, numerical_embeddings])

        outputs = encoder(feature_embeddings)

        if head is not None:
            outputs = head(outputs)

        super().__init__(inputs=inputs,
                         outputs=outputs,
                         **kwargs)

        self.input_dim = input_dim
        self.encoder = encoder
        self.categorical_feature_embedding = categorical_feature_embedding
        self.numerical_feature_embedding = numerical_feature_embedding
        self.head = head

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'categorical_feature_embedding': keras.layers.serialize(self.categorical_feature_embedding),
                       'numerical_feature_embedding': keras.layers.serialize(self.numerical_feature_embedding),
                       'encoder': keras.layers.serialize(self.encoder),
                       'head': keras.layers.serialize(self.head),
                       })
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        encoder = keras.layers.deserialize(config.pop("encoder"))
        categorical_feature_embedding = keras.layers.deserialize(config.pop("categorical_feature_embedding"))
        numerical_feature_embedding = keras.layers.deserialize(config.pop("numerical_feature_embedding"))
        head = keras.layers.deserialize(config.pop("head"))
        return cls(input_dim=input_dim,
                   categorical_feature_embedding=categorical_feature_embedding,
                   numerical_feature_embedding=numerical_feature_embedding,
                   encoder=encoder,
                   head=head,
                   **config)


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class SAINTPretrainer(keras.Model):
    """
    SAINTPretrainer model with LayerFlow design.
    It is based on the pretraining architecture of the SAINT model
    proposed by Gowthami Somepalli et al.
    in the paper,
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.

    SAINT performs attention over both rows and columns.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        model: ``keras.Model``,
            An instance of the ``SAINT`` model that is to be pretrained.

        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        mixup: ``keras.layers.Layer``,
            An instance of ``MixUp`` layer or any custom layer that can work
            in its place.
            You can import the ``MixUp`` layer as follows,
                >>> from teras.layers import MixUp

        cutmix: ``keras.layers.Layer``,
            An instance of ``CutMix`` layer or any custom layer that can work
            in its place.
            You can import the `CutMix` layer as follows,
                >>> from teras.layers import CutMix

        projection_head_1: ``keras.layers.Layer``,
            An instance of ``SAINTProjectionHead`` layer that is used to project embeddings
            of *real* samples to a lower dimensionality before reconstructing the inputs.
            You can import the ``SAINTProjectionHead`` layer as follows,
                >>> from teras.layerflow.layers import SAINTProjectionHead

        projection_head_2: ``keras.layers.Layer``,
            An instance of ``SAINTProjectionHead`` layer that is used to project embeddings
            of *augmented* samples to a lower dimensionality before reconstructing the inputs.
            You can import the ``SAINTProjectionHead`` layer as follows,
                >>> from teras.layerflow.layers import SAINTProjectionHead

        reconstruction_head: ``keras.layers.Layer``,
            An instance of ``SAINTReconstructionHead`` which applies a separate ``ReconstructionHeadBlock``
            to reconstruct the input features.
            You can import the ``SAINTReconstructionHead`` as follows,
                >>> from teras.layerflow.layers import SAINTReconstructionHead

        temperature: ``float``, default 0.7,
            Temperature value used in the computation of the InfoNCE contrastive loss.

        lambda_: ``float``, default 10,
            Controls the weightage of denoising loss in the summation of denoising and
            contrastive loss.
    """
    def __init__(self,
                 model: keras.Model,
                 features_metadata: dict,
                 mixup: keras.layers.Layer = None,
                 cutmix: keras.layers.Layer = None,
                 projection_head_1: keras.layers.Layer = None,
                 projection_head_2: keras.layers.Layer = None,
                 reconstruction_head: keras.layers.Layer = None,
                 temperature: float = 0.7,
                 lambda_: float = 10.,
                 **kwargs):
        super().__init__(**kwargs)
        self.model = model
        if mixup is None:
            raise ValueError("`mixup` cannot be None. Please pass an instance of `MixUp` layer. "
                             "You can import it as, `from teras.layers import MixUp`")

        if cutmix is None:
            raise ValueError("`cutmix` cannot be None. Please pass an instance of `CutMix` layer. "
                             "You can import it as, `from teras.layers import CutMix``")

        if projection_head_1 is None:
            raise ValueError("`projection_head_1` cannot be None. "
                             "Please pass an instance of `SAINTProjectionHead` layer. "
                             "You can import it as, `from teras.layers import SAINTProjectionHead``")

        if projection_head_2 is None:
            raise ValueError("`projection_head_2` cannot be None. "
                             "Please pass an instance of `SAINTProjectionHead` layer. "
                             "You can import it as, `from teras.layers import SAINTProjectionHead`")

        if reconstruction_head is None:
            raise ValueError("`reconstruction_head` cannot be None. "
                             "Please pass an instance of `SAINTReconstructionHead` layer. "
                             "You can import it as, `from teras.layers import SAINTReconstructionHead`")

        self.features_metadata = features_metadata
        self._categorical_features_exist = len(self.features_metadata["categorical"]) > 0
        self._numerical_features_exist = len(self.features_metadata["numerical"]) > 0

        self.mixup = mixup
        self.cutmix = cutmix
        self.projection_head_1 = projection_head_1
        self.projection_head_2 = projection_head_2
        self.reconstruction_head = reconstruction_head
        self.temperature = temperature
        self.lambda_ = lambda_

        self.contrastive_loss_tracker = keras.metrics.Mean(name="contrastive_loss")
        self.denoising_loss_tracker = keras.metrics.Mean(name="denoising_loss")

    def get_pretrained_model(self):
        """Returns pretrained model"""
        return self.model

    @property
    def pretrained_model(self):
        """Returns pretrained model"""
        return self.model

    def compile(self,
                contrastive_loss=info_nce_loss,
                denoising_loss=denoising_loss,
                **kwargs):
        super().compile(**kwargs)
        self.contrastive_loss = contrastive_loss
        self.denoising_loss = denoising_loss

    def call(self, inputs):
        x = inputs

        # Apply cutmix on the raw input space
        x_prime = self.cutmix(x)

        # Embed the raw inputs as well as cutmixed data
        p = None
        p_prime = None
        if self._categorical_features_exist:
            p = self.model.categorical_feature_embedding(x)
            p_prime = self.model.categorical_feature_embedding(x_prime)

        if self._numerical_features_exist:
            numerical_features = self.model.numerical_feature_embedding(x)
            numerical_features_prime = self.model.numerical_feature_embedding(x_prime)
            if p is not None:
                p = tf.concat([p, numerical_features],
                              axis=1)
                p_prime = tf.concat([p_prime, numerical_features_prime],
                                    axis=1)
            else:
                p = numerical_features
                p_prime = numerical_features_prime

        # Apply mixup on the embedding space -- only to the augment data
        p_prime = self.mixup(p_prime)

        # Pass these embeddings through saint encoder
        r = self.model.encoder(p)
        r_prime = self.model.encoder(p_prime)

        # Pass the encoded features through projection heads
        z = self.projection_head_1(r)
        z_prime = self.projection_head_2(r_prime)
        # Normalize
        z = z / tf.norm(z, axis=-1, keepdims=True)
        z_prime = z_prime / tf.norm(z_prime, axis=-1, keepdims=True)
        # Flatten last two dimensions
        z = tf.reshape(z, shape=(tf.shape(z)[0], tf.reduce_prod(tf.shape(z)[1:])))
        z_prime = tf.reshape(z_prime, shape=(tf.shape(z_prime)[0], tf.reduce_prod(tf.shape(z_prime)[1:])))

        # To reconstruct the input data, we pass the encodings of augmented data
        # i.e. the `r_prime` through a reconstruction head
        reconstructed_samples = self.reconstruction_head(r_prime)

        return z, z_prime, reconstructed_samples

    def train_step(self, data):
        if isinstance(data, tuple):
            data = data[0]

        with tf.GradientTape() as tape:
            z, z_prime, reconstructed_samples = self(data)
            c_loss = self.contrastive_loss(real_projection_outputs=z,
                                           augmented_projection_outputs=z_prime,
                                           temperature=self.temperature)
            d_loss = self.denoising_loss(real_samples=data,
                                         reconstructed_samples=reconstructed_samples,
                                         categorical_features_metadata=self.features_metadata["categorical"])

            loss = c_loss + self.lambda_ * d_loss
        gradients = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_weights))

        self.contrastive_loss_tracker.update_state(c_loss)
        self.denoising_loss_tracker.update_state(d_loss)
        # If user has passed any additional metrics to compile, we should update their states
        if len(self.compiled_metrics.metrics) > 0:
            self.compiled_metrics.update_state(data, reconstructed_samples)
        # If user has passed any additional losses to compile, we should call them
        if self.compiled_loss._losses is not None:
            self.compiled_loss(data, reconstructed_samples)
        results = {m.name: m.result() for m in self.metrics}
        return results

    def get_config(self):
        config = super().get_config()
        new_config = {'model': keras.layers.serialize(self.model),
                      'features_metadata': self.features_metadata,
                      'mixup': keras.layers.serialize(self.mixup),
                      'cutmix': keras.layers.serialize(self.cutmix),
                      'projection_head_1': keras.layers.serialize(self.projection_head_1),
                      'projection_head_2': keras.layers.serialize(self.projection_head_2),
                      'reconstruction_head': keras.layers.serialize(self.reconstruction_head),
                      'temperature': self.temperature,
                      'lambda_': self.lambda_,
                      }
        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        features_metadata = config.pop("features_metadata")
        mixup = keras.layers.deserialize(config.pop("mixup"))
        cutmix = keras.layers.deserialize(config.pop("cutmix"))
        projection_head_1 = keras.layers.deserialize(config.pop("projection_head_1"))
        projection_head_2 = keras.layers.deserialize(config.pop("projection_head_2"))
        reconstruction_head = keras.layers.deserialize(config.pop("reconstruction_head"))
        return cls(model=model,
                   features_metadata=features_metadata,
                   mixup=mixup,
                   cutmix=cutmix,
                   projection_head_1=projection_head_1,
                   projection_head_2=projection_head_2,
                   reconstruction_head=reconstruction_head,
                   **config)



import tensorflow as tf
from teras.layers.saint.saint_reconstruction_head import SAINTReconstructionHead
from teras.utils import get_features_metadata_for_embedding
import pandas as pd
import numpy as np


def test_saint_reconstruction_head_output_shape():
    data = pd.DataFrame({"income": np.ones(10),
                         "goals": np.ones(10),
                         'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                         'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
    categorical_feats = ["player_level", "shirt_number"]
    numerical_feats = ["income", "goals"]
    features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                            categorical_features=categorical_feats,
                                                            numerical_features=numerical_feats)
    sum_of_features_cardinalities = sum(map(lambda x: len(features_metadata["categorical"][x][1]),
                                            features_metadata["categorical"])
                                        ) + len(numerical_feats)

    saint_reconstruction_head = SAINTReconstructionHead(features_metadata,
                                                        embedding_dim=32)
    inputs = tf.ones(shape=(16, 4, 32), dtype=tf.float32)
    outputs = saint_reconstruction_head(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 16
    assert tf.shape(outputs)[1] == sum_of_features_cardinalities



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.vime")
class VimeMaskGenerationAndCorruption(keras.layers.Layer):
    """
    A stateless layer that generates masks and corrupted samples
    By implementing this functionality in a layer's call method, we can use this layer in a sequential or functional
    manner and this way the layer and hence the generator functions it calls will have access to the batch size dimension
    because we'll precede this layer with an input layer that specifies the batch dimension and input shape in general
    and pack these both in a sequential model or plug them together using functional API

    Args:
        p_m: ``float``, default 0.3,
            Corruption probability. (Note: Don't do it in real life!)
    """
    def __init__(self,
                 p_m: float = 0.3,
                 **kwargs):
        super().__init__(**kwargs)
        self.p_m = p_m

    def call(self, inputs):
        """inputs: Batch of unlabeled X"""

        # Generate Mask Vector
        mask = tf.random.stateless_binomial(shape=tf.shape(inputs),
                                            seed=(0, 0),
                                            counts=1,
                                            probs=self.p_m,
                                            output_dtype=tf.float32)

        # Generate corrupted samples
        num_samples = tf.shape(inputs)[0]
        dim = tf.shape(inputs)[1]

        x_bar = tf.TensorArray(size=dim, dtype=tf.float32)
        for i in range(dim):
            idx = tf.random.shuffle(tf.range(num_samples))
            x_bar = x_bar.write(i, tf.gather(inputs[:, i], idx))
        x_bar = tf.transpose(x_bar.stack())

        # Corrupt Samples
        x_tilde = inputs * (1 - mask) + x_bar * mask
        return x_tilde

    def get_config(self):
        config = super().get_config()
        config.update({'p_m': self.p_m,
                       })
        return config



from tensorflow import keras
from teras.layerflow.layers.common.common import HiLOL


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.saint")
class SAINTProjectionHead(HiLOL):
    """
    SAINTProjectionHead layer with LayerFlow design.
    It is used in the contrastive learning phase of
    the ``SAINTPretrainer`` to project embeddings to a lower dimension.
    According to the SAINT paper,
    "The use of a projection head to reduce dimensionality before computing
    contrastive loss is common in vision and indeed also improves results
    on tabular data."

    Reference(s):
    https://arxiv.org/abs/2106.01342

    Args:
        hidden_block: ``keras.layers.Layer``,
            Hidden block to use in the projection head.
            It can be as simple as a single dense layer with "relu" activation,
            or as complex as you want.
            If the official implementation, the hidden dimensionality is
            computed as below,
            `hidden_dim = 6 * embedding_dim * number_of_features // 5`

        output_layer: ``keras.layers.Layer``,
            Output layer to use in the projection head.
            It should be a simple dense layer that project data to a lower dimension.
            If the official implementation, the output dimensionality is computed as
            below,
            `output_dim = embedding_dim * number_of_features // 5`
    """
    def __init__(self,
                 hidden_block: keras.layers.Layer,
                 output_layer: keras.layers.Layer,
                 **kwargs):
        super().__init__(hidden_block=hidden_block,
                         output_layer=output_layer,
                         **kwargs)


from tensorflow import keras
from teras.layerflow.layers.saint.saint_reconstruction_head import (SAINTReconstructionHeadBlock as _SAINTReconstructionHeadBlockLF,
                                                                    SAINTReconstructionHead as _SAINTReconstructionHeadLF)


@keras.saving.register_keras_serializable(package="teras.layers.saint")
class SAINTReconstructionHeadBlock(_SAINTReconstructionHeadBlockLF):
    """
    ReconstructionBlock layer that is used in constructing ReconstructionHead.
    One ReconstructionBlock is created for each feature in the dataset.
    The inputs to this layer are first mapped to hidden dimensions
    and then projected to the cardinality of the feature.

    Args:
        feature_cardinality: ``int``,
            Cardinality of the given input feature.
            For categorical features, it is equal to the number of classes
            in the feature, and for numerical features, it is equal to 1.

        hidden_dim: ``int``, default 32,
            Dimensionality of the hidden layer.

        hidden_activation: ``str``, default "relu",
            Activation function to use in the hidden layer.
    """
    def __init__(self,
                 feature_cardinality: int,
                 hidden_dim: int = 32,
                 hidden_activation="relu",
                 **kwargs):
        hidden_block = keras.layers.Dense(hidden_dim,
                                          activation=hidden_activation,
                                          name="hidden_block")
        output_layer = keras.layers.Dense(feature_cardinality,
                                          name="output_layer")
        super().__init__(hidden_block=hidden_block,
                         output_layer=output_layer,
                         **kwargs)
        self.feature_cardinality = feature_cardinality
        self.hidden_dim = hidden_dim
        self.hidden_activation = hidden_activation

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'feature_cardinality': self.feature_cardinality,
                  'hidden_dim': self.hidden_dim,
                  'hidden_activation': self.hidden_activation,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        feature_cardinality = config.pop("feature_cardinality")
        return cls(feature_cardinality, **config)


@keras.saving.register_keras_serializable(package="teras.layers.saint")
class SAINTReconstructionHead(_SAINTReconstructionHeadLF):
    """
        ReconstructionHead layer for ``SAINTPretrainer`` model.
    SAINT applies a separate single hidden layer MLP block
    (here we name it, the reconstruction block)
    with an output layer where output dimensions are equal
    to the number of categories in the case of categorical
    features and 1 in the case of numerical features.

    Args:
        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                                                                        numerical_features,
                                                                        categorical_features)
        embedding_dim: ``int``, default 32,
            Embedding dimensions being used in the pretraining model.
            Used in the computation of the hidden dimensions for each
            reconstruction (mlp) block for each feature.
    """
    def __init__(self,
                 features_metadata: dict,
                 embedding_dim: int = 32,
                 **kwargs):
        categorical_features_metadata = features_metadata["categorical"]
        numerical_features_metadata = features_metadata["numerical"]
        num_categorical_features = len(categorical_features_metadata)
        num_numerical_features = len(numerical_features_metadata)

        # feature_cardinalities: Dimensions of each feature in the input
        # For a categorical feature, it is equal to the number of unique categories in the feature
        # For a numerical features, it is equal to 1
        features_cardinalities = []
        # recall that categorical_features_metadata dict maps feature names to a tuple of
        # feature id and unique values in the feature
        if num_categorical_features > 0:
            features_cardinalities = list(map(lambda x: len(x[1]), categorical_features_metadata.values()))
        features_cardinalities.extend([1] * num_numerical_features)

        # For the computation of denoising loss, we use a separate MLP block for each feature
        # we call the combined blocks, reconstruction heads
        reconstruction_blocks = [
            SAINTReconstructionHeadBlock(feature_cardinality=cardinality,
                                         hidden_dim=embedding_dim * 5,
                                         )
            for cardinality in features_cardinalities]

        super().__init__(reconstruction_blocks, **kwargs)

        self.features_metadata = features_metadata
        self.embedding_dim = embedding_dim

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'features_metadata': self.features_metadata,
                  'embedding_dim': self.embedding_dim,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        # features_metadata is the only positional argument
        features_metadata = config.pop("features_metadata")
        return cls(features_metadata, **config)



from tensorflow import keras
from teras.utils import get_activation
from teras.layerflow.layers.common.transformer import (FeedForward as _FeedForwardLF,
                                                       Transformer as _TransformerLF,
                                                       Encoder as _EncoderLF)
from teras.utils.types import ActivationType


@keras.saving.register_keras_serializable(package="teras.layers.common")
class FeedForward(_FeedForwardLF):
    """
    FeedForward layer that is used in the ``Transformer`` layer of the ``Encoder`` block
    of Transformer architecture.
    It is shared by all transformer based architectures for tabular data.
    The first layer expands the inputs to `embedding dimensions` times `multiplier`
    and the second layer projects them back to the embedding dimensions.
    Usually, since the feed forward layer follows the ``MultiHeadAttention`` layer
    whose outputs are of `embedding dimensions` so it essentially takes
    MultiHeadAttention layer's outputs as inputs, expands them to input dimensions
    times the multiplier and then projects them back to the input dimensions.

    Args:
        embedding_dim: `int`, default 16,
            Embedding dimensions used in the given
            architecture to usually embedded the numerical features but at times
            to also embedd the numerical features like in SAINT architecture.

        multiplier: ``int``, default 4,
            Multiplier that is multipled with embedding dims
            and the resultant value is used as hidden dimensions.

        dropout: ``float``, default 0.,
            Dropout rate to use in the dropout layer that is applied after
            the hidden layer.

        activation: default "geglu",
            Activation function to use in the hidden layer.
            By default, ``geglu`` activation is used.
            You can import it as,
                >>> from teras.activations import geglu
    """
    def __init__(self,
                 embedding_dim: int = 32,
                 multiplier: int = 4,
                 dropout: float = 0.,
                 activation: ActivationType = "geglu",
                 **kwargs):
        hidden_block = keras.models.Sequential(name="feed_forward_hidden_block")
        hidden_block.add(keras.layers.Dense(embedding_dim * multiplier,
                                            activation=get_activation(activation)))
        hidden_block.add(keras.layers.Dropout(dropout))
        output_layer = keras.layers.Dense(embedding_dim)

        super().__init__(hidden_block=hidden_block,
                         output_layer=output_layer,
                         **kwargs)

        self.embedding_dim = embedding_dim
        self.multiplier = multiplier
        self.dropout = dropout
        self.activation = activation

    def get_config(self):
        activation_serialized = self.activation
        if not isinstance(activation_serialized, str):
            activation_serialized = keras.layers.serialize(activation_serialized)
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'embedding_dim': self.embedding_dim,
                  'multiplier': self.multiplier,
                  'dropout': self.dropout,
                  'activation': activation_serialized}
        return config


@keras.saving.register_keras_serializable(package="teras.layers.common")
class Transformer(_TransformerLF):
    """
    Transformer layer that is used as a building block for
    constructing the ``Encoder`` block of transformer based architectures.
    It applies the ``MultiHeadAttention`` followed by the FeedForward layer
    and also employs residual connections.

    Args:
        num_attention_heads: ``int``, default 8,
            Number of heads to use in the ``MultiHeadAttention`` layer.

        embedding_dim: ``int``, default 32,
            Embedding dimensions being used in the overall architecture.
            These serve as the `key dimensions` in the
            ``MultiHeadAttention`` layer and are multiplied by the
            ``mutliplier`` in the ``FeedForward`` layer to compute the
            hidden dimensions to project the inputs into.

        attention_dropout: ``float``, default 0.,
            Dropout rate for the ``MultiHeadAttention`` layer.

        feedforward_dropout: ``float``, default 0.,
            Dropout rate to use in for the dropout layer in the ``FeedForward`` layer.

        norm_epsilon: ``float``, default 1e-6,
            Normalization value to use for ``LayerNormalization`` layer.
    """
    def __init__(self,
                 embedding_dim: int = 32,
                 num_attention_heads: int = 8,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 feedforward_multiplier: int = 4,
                 norm_epsilon: float = 1e-6,
                 **kwagrs):
        multi_head_attention = keras.layers.MultiHeadAttention(
            num_heads=num_attention_heads,
            key_dim=embedding_dim,
            dropout=attention_dropout
        )
        layer_norm_1 = keras.layers.LayerNormalization(epsilon=norm_epsilon)
        feed_forward = FeedForward(embedding_dim,
                                   multiplier=feedforward_multiplier,
                                   dropout=feedforward_dropout)
        layer_norm_2 = keras.layers.LayerNormalization(epsilon=norm_epsilon)
        super().__init__(multi_head_attention=multi_head_attention,
                         feed_forward=feed_forward,
                         layer_norm_1=layer_norm_1,
                         layer_norm_2=layer_norm_2,
                         **kwagrs)
        self.embedding_dim = embedding_dim
        self.num_attention_heads = num_attention_heads
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.feedforward_multiplier = feedforward_multiplier
        self.norm_epsilon = norm_epsilon

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'embedding_dim': self.embedding_dim,
                  'num_attention_heads': self.num_attention_heads,
                  'attention_dropout': self.attention_dropout,
                  'feedforward_dropout': self.feedforward_dropout,
                  'feedforward_multiplier': self.feedforward_multiplier,
                  'norm_epsilon': self.norm_epsilon}
        return config


@keras.saving.register_keras_serializable(package="teras.layers.common")
class Encoder(_EncoderLF):
    """
    Encoder for transformer based architectures.
    It is simply a stack of `num_transformer_layers`
    that is applied to the inputs.

    Args:
        num_transformer_layer: ``int``, default 6,
            Number of transformer layers to use in the Encoder.

        num_attention_heads: ``int``, default 8,
            Number of heads to use in the ``MultiHeadAttention`` layer
            that is used to construct the ``Transformer`` layer.

        embedding_dim: ``int``, default 32,
            Embedding dimensions being used in the overall architecture.
            These serve as the `key dimensions` in the
            ``MultiHeadAttention`` layer and are multiplied by the
            ``mutliplier`` in the ``FeedForward`` layer to compute the
            hidden dimensions to project the inputs into.

        attention_dropout: ``float``, default 0.,
            Dropout rate for the ``MultiHeadAttention`` layer.

        feedforward_dropout: ``float``, default 0.,
            Dropout rate to use in for the dropout layer in the ``FeedForward`` layer.

        norm_epsilon: ``float``, default 1e-6,
            Normalization value to use for ``LayerNormalization`` layer.
    """
    def __init__(self,
                 num_transformer_layers: int = 6,
                 num_attention_heads: int = 8,
                 embedding_dim: int = 32,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 feedforward_multiplier: int = 4,
                 norm_epsilon: float = 1e-6,
                 **kwargs):
        transformer_layers = keras.models.Sequential(name="transformer_layers")
        for i in range(num_transformer_layers):
            transformer_layers.add(
                Transformer(num_attention_heads=num_attention_heads,
                            embedding_dim=embedding_dim,
                            attention_dropout=attention_dropout,
                            feedforward_dropout=feedforward_dropout,
                            feedforward_multiplier=feedforward_multiplier,
                            norm_epsilon=norm_epsilon,
                            name=f"transformer_layer_{i+1}"))
        super().__init__(transformer_layers,
                         **kwargs)
        self.num_transformer_layers = num_transformer_layers
        self.num_attention_heads = num_attention_heads
        self.embedding_dim = embedding_dim
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.feedforward_multiplier = feedforward_multiplier
        self.norm_epsilon = norm_epsilon

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'num_transformer_layers': self.num_transformer_layers,
                  'num_attention_heads': self.num_attention_heads,
                  'embedding_dim': self.embedding_dim,
                  'attention_dropout': self.attention_dropout,
                  'feedforward_dropout': self.feedforward_dropout,
                  'feedforward_multiplier': self.feedforward_multiplier,
                  'norm_epsilon': self.norm_epsilon}
        return config

    @classmethod
    def from_config(cls, config):
        # we need to override the from_config method because the parent
        # is layerflow version of Encoder which tried to extract
        # the value against `transformer_layers` key to deserialize
        # which causes KeyError
        return cls(**config)


# Copyright 2023 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# TabTransfomer models
from teras.layerflow.models.tabtransformer import (TabTransformer,
                                                   TabTransformerPretrainer
                                                   )

# TabNet models
from teras.layerflow.models.tabnet import (TabNet,
                                           TabNetPretrainer)


# SAINT models
from teras.layerflow.models.saint import (SAINT,
                                          SAINTPretrainer)


# FT Transformer models
from teras.layerflow.models.ft_transformer import FTTransformer


# DNFNet models
from teras.layerflow.models.dnfnet import DNFNet


# NODE models
from teras.layerflow.models.node import NODE


# RTDLResNet models
from teras.layerflow.models.rtdl_resnet import RTDLResNet


# GAIN models
from teras.layerflow.models.gain import GAIN


# PCGAIN models
from teras.layerflow.models.pcgain import PCGAIN


# CTGAN models
from teras.layerflow.models.ctgan import CTGAN


# TVAE models
from teras.layerflow.models.tvae import TVAE


# VIME models
from teras.layerflow.models.vime import (VimeSelf,
                                         VimeSemi)



from tensorflow import keras
from teras.layerflow.layers.common.common import HiLOL
from typing import List

LIST_OF_LAYERS = List[keras.layers.Layer]


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.common")
class FeedForward(HiLOL):
    """
    FeedForward layer with LayerFlow design.
    It is used in the Transformer layer of the ``Encoder`` layer
    of Transformer architecture.
    It is shared by all transformer based architectures for tabular data.

    Args:
        hidden_block: ``keras.layers.Layer``,
            A dense layer or any custom layer that can serve as the hidden
            block for the ``FeedForward`` layer.

        output_layer: ``keras.layers.Layer``,
            A dense layer or any custom layer that projects data back to
            the input/embedding dimensions. Since the ``FeedForward`` receives
            feature embeddings as inputs, it after apply hidden block,
            projects them back to the embedding dimensions.
    """
    def __init__(self,
                 hidden_block: keras.layers.Layer,
                 output_layer: keras.layers.Layer,
                 **kwargs):
        super().__init__(hidden_block=hidden_block,
                         output_layer=output_layer,
                         **kwargs)


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.common")
class Transformer(keras.layers.Layer):
    """
    Transformer layer with LayerFlow design.
    It is used to made up the ``Encoder`` layer of the given
    Transformer-based architecture.

    Args:
        multi_head_attention: ``keras.layers.Layer``,
            An instance of ``MultiHeadAttention`` layer or any layer
            that can be sued in place of ``MultiHeadAttention`` layer.
            You can import this layer as follows,
                >>> from keras.layers import MultiHeadAttention

        feed_forward: ``keras.layers.Layer``,
            An instance of ``FeedForward`` layer or any layer that
            can be used in its place.
            You can import the ``FeedForward`` layer as follows,
                >>> from teras.layerflow.layers import FeedForward

        layer_norm_1: ``keras.layers.Layer``,
            An instance of ``LayerNormalization`` instance applied after
            first skip connection.
            If None, a default ``LayerNormalization`` instance is used.
            You can import this layer as follows,
                >>> from keras.layers import LayerNormalization

        layer_norm_1: ``keras.layers.Layer``, optional,
            An instance of ``LayerNormalization`` instance applied after
            second skip connection.
            If None, a default ``LayerNormalization`` instance is used.
            You can import this layer as follows,
                >>> from keras.layers import LayerNormalization
    """
    def __init__(self,
                 multi_head_attention: keras.layers.Layer,
                 feed_forward: keras.layers.Layer,
                 layer_norm_1: keras.layers.Layer = None,
                 layer_norm_2: keras.layers.Layer = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.multi_head_attention = multi_head_attention
        self.skip_1 = keras.layers.Add(name="skip_1")
        self.layer_norm_1 = layer_norm_1
        if self.layer_norm_1 is None:
            self.layer_norm_1 = keras.layers.LayerNormalization()
        self.feed_forward = feed_forward
        self.skip_2 = keras.layers.Add(name="skip_2")
        self.layer_norm_2 = layer_norm_2
        if self.layer_norm_2 is None:
            self.layer_norm_2 = keras.layers.LayerNormalization()

    def call(self, inputs):
        attention_out = self.multi_head_attention(inputs, inputs)
        x = self.skip_1([attention_out, inputs])
        x = self.layer_norm_1(x)
        feedforward_out = self.feed_forward(x)
        x = self.skip_2([feedforward_out, x])
        x = self.layer_norm_2(x)
        return x

    def get_config(self):
        config = super().get_config()
        new_config = {'multi_head_attention': keras.layers.serialize(self.multi_head_attention),
                      'feed_forward': keras.layers.serialize(self.feed_forward),
                      'layer_norm_1': keras.layers.deserialize(self.layer_norm_1),
                      'layer_norm_2': keras.layers.deserialize(self.layer_norm_2)
                      }
        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        multi_head_attention = keras.layers.deserialize(config.pop("multi_head_attention"))
        feed_forward = keras.layers.deserialize(config.pop("feed_forward"))
        transformer = keras.layers.deserialize(config.pop("transformer"))
        layer_norm_1 = keras.layers.deserialize(config.pop("layer_norm_1"))
        layer_norm_2 = keras.layers.deserialize(config.pop("layer_norm_2"))
        return cls(multi_head_attention=multi_head_attention,
                   feed_forward=feed_forward,
                   transformer=transformer,
                   layer_norm_1=layer_norm_1,
                   layer_norm_2=layer_norm_2,
                   **config)


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.common")
class Encoder(keras.layers.Layer):
    """
    Encoder layer with LayerFlow design.
    It is made up of several ``Transformer`` layers and encodes the
    input embeddings to useful representations that can then be used
    for the task at hand, such as classificaiton or regression.

    Args:
        transformer_layers: ``List[keras.layers.Layer]``,
            A list of ``Transformer`` layers that make up the encoder
            layer.
            It can also be an instance of Keras Model made up of
            ``Transformer`` or any other layers in the mix.
            You can import the ``Transformer`` layer as follows,
                >>> from teras.layerflow.layers import Transformer
    """
    def __init__(self,
                 transformer_layers: LIST_OF_LAYERS,
                 **kwargs):
        super().__init__(**kwargs)
        self.transformer_layers = transformer_layers
        if isinstance(self.transformer_layers, (list, tuple)):
            self.transformer_layers = keras.models.Sequential(
                self.transformer_layers,
                name="transformer_layers"
            )

    def call(self, inputs):
        outputs = self.transformer_layers(inputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({'transformer_layers': keras.layers.serialize(self.transformer_layers)})
        return config

    @classmethod
    def from_config(cls, config):
        transformer_layers = keras.layers.deserialize(config.pop("transformer_layers"))
        return cls(transformer_layers, **config)



import tensorflow as tf
from tensorflow import keras
from teras.losses import VimeSelfSupervisedLoss


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class VimeSelf(keras.Model):
    """
    Self-supervised learning part of the VIME architecture,
    proposed by Jinsung Yoon et al. in the paper
    "VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain"

    Reference(s):
        https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html

    Args:
        input_dim: ``int``,
            The dimensionality of the input dataset,
            or the number of features in the dataset.

        encoder: ``keras.layers.Layer``,
            An instance of the ``VimeEncoder`` layer or any custom
            layer that can work in its place.
            You can import the ``VimeEncoder`` layer as follows,
                >>> from teras.layers import VimeEncoder

        mask_estimator: ``keras.layers.Layer``,
            An instance of the ``VimeMaskEstimator`` layer or any custom
            layer that can work in its place.
            You can import the ``VimeMaskEstimator`` layer as follows,
                >>> from teras.layers import VimeMaskEstimator

        feature_estimator: ``keras.layers.Layer``,
            An instance of the ``VimeFeatureEstimator`` layer or any custom
            layer that can work in its place.
            You can import the ``VimeFeatureEstimator`` layer as follows,
                >>> from teras.layers import VimeFeatureEstimator
    """
    def __init__(self,
                 input_dim: int,
                 encoder: keras.layers.Layer = None,
                 mask_estimator: keras.layers.Layer = None,
                 feature_estimator: keras.layers.Layer = None,
                 **kwargs
                 ):

        if encoder is None:
            raise ValueError("`encoder` cannot be None. You must pass an instance of `VimeEncoder` or any custom layer "
                             "that can work in its place. \n"
                             "You can import the `VimeEncoder` as `from teras.layers import VimeEncoder`")

        if mask_estimator is None:
            raise ValueError("`mask_estimator` cannot be None. You must pass an instance of `VimeMaskEstimator` or any "
                             "custom layer that can work in its place. \n"
                             "You can import the `VimeMaskEstimator` as `from teras.layers import VimeMaskEstimator`")

        if feature_estimator is None:
            raise ValueError("`feature_estimator` cannot be None. You must pass an instance of `VimeFeatureEstimator` "
                             "or any custom layer that can work in its place. \n"
                             "You can import the `VimeFeatureEstimator` as "
                             "`from teras.layers import VimeFeatureEstimator`")

        inputs = keras.layers.Input(shape=(input_dim,))
        # Encoding using VimeEncoder
        encoded_inputs = encoder(inputs)
        # Mask estimator
        mask_out = mask_estimator(encoded_inputs)
        # Feature estimator
        feature_out = feature_estimator(encoded_inputs)
        # Calling model.get_layer("encode") returns error that the encoder layer is not connected
        # to any node. To combat that issue, instead of calling layers one by one on inputs
        # directly, we instead make a functional model and call it on inputs.
        # Making a functional model allows us to create a Input layer and connect it with
        # the encoder layer.
        super().__init__(inputs=inputs,
                         outputs={"mask_estimator": mask_out,
                                  "feature_estimator": feature_out},
                         **kwargs)
        self.input_dim = input_dim
        self.encoder = encoder
        self.mask_estimator = mask_estimator
        self.feature_estimator = feature_estimator

    def compile(self,
                loss={"mask_estimator": keras.losses.MeanSquaredError(),
                      "feature_estimator": keras.losses.BinaryCrossentropy()},
                loss_weights={"mask_estimator": 1.0,
                              "feature_estimator": 2.0},
                **kwargs):
        super().compile(loss=loss,
                        loss_weights=loss_weights,
                        **kwargs)

    def get_encoder(self):
        """
        Retrieves the encoder part of the trained model
        Returns:
            Encoder part of the model
        """
        # Extract encoder part
        encoder = keras.models.Model(inputs=self.input, outputs=self.encoder.output)
        return encoder

    def get_config(self):
        config = super().get_config()
        config.update({"input_dim": self.input_dim,
                       "encoder": keras.layers.serialize(self.encoder),
                       "mask_estimator": keras.layers.serialize(self.mask_estimator),
                       "feature_estimator": keras.layers.serialize(self.feature_estimator)
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        encoder = keras.layers.deserialize(config.pop("encoder"))
        mask_estimator = keras.layers.deserialize(config.pop("mask_estimator"))
        feature_estimator = keras.layers.deserialize(config.pop("feature_estimator"))
        return cls(input_dim=input_dim,
                   encoder=encoder,
                   mask_estimator=mask_estimator,
                   feature_estimator=feature_estimator,
                   **config)


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class VimeSemi(keras.Model):
    """
    Semi-supervised learning part of the VIME architecture,
    proposed by Jinsung Yoon et al. in the paper
    "VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain"

    Reference(s):
        https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset.

        pretrained_encoder: ``keras.Model``,
            Pretrained instance of ``VimEncoder``, pretrained during the 
            Self-Supervised learning phase.
            You can access it after training the ``VimeSelf`` model through its
            ``.get_encoder()`` method.

        mask_generation_and_corruption: ``keras.layers.Layer``,
            An instance oof the ``VimeMaskGenerationAndCorruption`` layer,
            or any custom layer that can work in its place.
            You can import the ``VimeMaskGenerationAndCorruption`` layer as follows,
                >>> from teras.layers import VimeMaskGenerationAndCorruption

        predictor: ``keras.layers.Layer``,
            An instance oof the ``VimePredictor`` layer,
            or any custom layer that can work in its place.
            You can import the ``VimePredictor`` layer as follows,
                >>> from teras.layers import VimePredictor

        K: ``int``, default 3,
            Number of augmented samples

        beta: ``float``, default 1.0,
            Hyperparameter to control supervised and unsupervised loss
    """
    def __init__(self,
                 input_dim: int,
                 pretrained_encoder: keras.Model = None,
                 mask_generation_and_corruption: keras.layers.Layer = None,
                 predictor: keras.layers.Layer = None,
                 K: int = 3,
                 beta: float = 1.0,
                 **kwargs):
        if pretrained_encoder is None:
            raise ValueError("`pretrained_encoder` cannot be None. You must pass an instance of pretrained encoder "
                             "trained during the self supervised phase. \n"
                             "You can access it after training the `VimeSelf` model through its "
                             "`.get_encoder()` method.")

        if mask_generation_and_corruption is None:
            raise ValueError("`mask_generation_and_corruption` cannot be None. You must pass an instance of "
                             "`VimeMaskGenerationAndCorruption` layer or any "
                             "custom layer that can work in its place. \n"
                             "You can import the `VimeMaskGenerationAndCorruption` as "
                             "`from teras.layers import VimeMaskGenerationAndCorruption`")

        if predictor is None:
            raise ValueError("`predictor` cannot be None. You must pass an instance of `VimePredictor` or any "
                             "custom layer that can work in its place. \n"
                             "You can import the `VimePredictor` as `from teras.layers import VimePredictor`")

        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.pretrained_encoder = pretrained_encoder
        self.K = K
        self.beta = beta

        self.mask_corrupt_block = keras.models.Sequential()
        self.input_layer = keras.layers.Input(shape=(self.input_dim,))
        self.mask_corrupt_block.add(self.input_layer)
        self.mask_generation_and_corruption = mask_generation_and_corruption
        self.mask_corrupt_block.add(self.mask_and_corrupt)

        self.predictor = predictor
        self.self_supervised_loss = VimeSelfSupervisedLoss()
        self.add_loss(self.self_supervised_loss)

    def train_step(self, data):
        labeled_dataset, unlabeled_dataset = data
        X_batch = labeled_dataset["x_labeled"]
        targets = labeled_dataset["y_labeled"]
        X_unlabeled_batch = unlabeled_dataset["x_unlabeled"]

        # Encode labeled data
        X_batch_encoded = self.pretrained_encoder(X_batch,
                                                  training=False)

        X_unlabeled_batch_encoded = list()
        for _ in range(self.K):
            X_unlabeled_batch_temp = self.mask_corrupt_block(X_unlabeled_batch)

            # Encode Corrupted Samples
            X_unlabeled_batch_temp = self.pretrained_encoder(X_unlabeled_batch_temp,
                                                             training=False)
            X_unlabeled_batch_encoded = X_unlabeled_batch_encoded + [X_unlabeled_batch_temp]

        # Convert list to tensor matrix
        X_unlabeled_batch_encoded = tf.stack(X_unlabeled_batch_encoded, axis=0)

        with tf.GradientTape() as tape:
            outputs = self({"labeled": X_batch_encoded,
                            "unlabeled": X_unlabeled_batch_encoded},
                           training=True)
            yu_loss = self.self_supervised_loss(y_true=None, y_pred=outputs['yv_hat_logit'])
            y_loss = self.compiled_loss(targets, outputs['y_hat_logit'])
            loss = y_loss + self.beta * yu_loss
        gradients = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_weights))
        self.compiled_metrics.update_state(targets, outputs['y_hat'])
        return {m.name: m.result() for m in self.metrics}

    def test_step(self, data):
        # Since at the time of prediction / evaluation we won't be processing any unlabeled data
        # so we get rid of all the unlabeled data logic / processing
        inputs, targets = data
        X_batch = inputs
        # Encode labeled data
        X_batch_encoded = self.pretrained_encoder(X_batch,
                                                  training=False)
        outputs = self({"labeled": X_batch_encoded},
                       training=False)
        self.compiled_loss(targets, outputs['y_hat_logit'])
        self.compiled_metrics.update_state(targets, outputs['y_hat'])
        return {m.name: m.result() for m in self.metrics}

    def call(self, inputs, training=None):
        X = inputs["labeled"]
        outputs = {}
        y_hat_logit, y_hat = self.predictor(X)
        outputs['y_hat_logit'] = y_hat_logit
        outputs['y_hat'] = y_hat
        # We process unlabeled data only while training
        if training:
            X_unlabled = inputs["unlabeled"]
            yv_hat_logit, yv_hat = self.predictor(X_unlabled)
            outputs['yv_hat_logit'] = yv_hat_logit
            outputs['yv_hat'] = yv_hat
        outputs['beta'] = self.beta
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'pretrained_encoder': keras.layers.serialize(self.pretrained_encoder),
                       'mask_generation_and_corruption': keras.layers.serialize(self.mask_generation_and_corruption),
                       'predictor': keras.layers.serialize(self.predictor),
                       'K': self.K,
                       'beta': self.beta,
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        pretrained_encoder = keras.layers.deserialize(config.pop("pretrained_encoder"))
        mask_generation_and_corruption = keras.layers.deserialize(config.pop("mask_generation_and_corruption"))
        predictor = keras.layers.deserialize(config.pop("predictor"))
        return cls(input_dim=input_dim,
                   pretrained_encoder=pretrained_encoder,
                   mask_generation_and_corruption=mask_generation_and_corruption,
                   predictor=predictor,
                   **config)



import tensorflow as tf
from tensorflow import keras
from teras.activations import gumbel_softmax


@keras.saving.register_keras_serializable(package="teras.layers.activation")
class GumbelSoftmax(keras.layers.Layer):
    """
    Implementation of the Gumbel Softmax activation
    proposed by Eric Jang et al. in the paper
    Categorical Reparameterization with Gumbel-Softmax

    Reference(s):
        https://arxiv.org/abs/1611.01144

    Args:
        temperature: ``float``, default 0.2,
            Controls the sharpness or smoothness of the resulting probability distribution.
            A higher temperature value leads to a smoother and more uniform probability distribution.
            Conversely, a lower temperature value makes the distribution concentrated around
            the category with the highest probability.

        hard: ``bool``, default False,
            Whether to return soft probabilities or hard one hot vectors.
    """
    def __init__(self,
                 temperature: float = 0.2,
                 hard: bool = False,
                 **kwargs):
        super().__init__(**kwargs)
        self.temperature = temperature
        self.hard = hard

    def call(self, logits: tf.Tensor):
        return gumbel_softmax(logits,
                              temperature=self.temperature,
                              hard=self.hard)

    def get_config(self):
        config = super().get_config()
        new_config = {'temperature': self.temperature,
                      'hard': self.hard}
        config.update(new_config)
        return config



import tensorflow as tf
from teras.layers.tabnet.tabnet_feature_transformer import TabNetFeatureTransformer


def test_tabnet_feature_transformer_valid_call():
    TabNetFeatureTransformer.reset_shared_layers()
    feature_transformer = TabNetFeatureTransformer(units=32)
    inputs = tf.ones((8, 10))
    outputs = feature_transformer(inputs)


def test_tabnet_feature_transformer_output_shape():
    TabNetFeatureTransformer.reset_shared_layers()
    feature_transformer = TabNetFeatureTransformer(units=32)
    inputs = tf.ones((8, 10))
    outputs = feature_transformer(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 32



import tensorflow as tf
from teras.layers.vime.vime_predictor import VimePredictor


def test_vime_predictor_valid_call():
    inputs = tf.ones((8, 5))
    predictor = VimePredictor(num_labels=2,
                              hidden_dim=16)
    y_hat_logit, y_hat = predictor(inputs)


def test_vime_predictor_output_shape():
    inputs = tf.ones((8, 5))
    predictor = VimePredictor(num_labels=2,
                              hidden_dim=16)
    y_hat_logit, y_hat = predictor(inputs)

    assert len(tf.shape(y_hat_logit)) == 2
    assert tf.shape(y_hat_logit)[0] == 8
    assert tf.shape(y_hat_logit)[1] == 2

    assert len(tf.shape(y_hat)) == 2
    assert tf.shape(y_hat)[0] == 8
    assert tf.shape(y_hat)[1] == 2



from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.ctgan")
class CTGANGeneratorBlock(keras.layers.Layer):
    """
    Residual Block for Generator as used by the authors of CTGAN
    proposed in the paper Modeling Tabular data using Conditional GAN.

    ``outputs = Concat([ReLU(BatchNorm(Dense(inputs))), inputs])``

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        units: ``int``, default 256,
            Dimensionality of the hidden layer
    """
    def __init__(self,
                 units: int = 256,
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.dense = keras.layers.Dense(self.units)
        self.batch_norm = keras.layers.BatchNormalization()
        self.relu = keras.layers.ReLU()
        self.concat = keras.layers.Concatenate(axis=1)

    def call(self, inputs):
        x = self.dense(inputs)
        x = self.batch_norm(x)
        x = self.relu(x)
        out = self.concat([x, inputs])
        return out

    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config



from tensorflow import keras


@keras.saving.register_keras_serializable("teras.layerflow.models")
class FTTransformer(keras.Model):
    """
    FTTransformer architecture with LayrFlow design.
    FT-Transformer is proposed by Yury Gorishniy et al.
    in the paper Revisiting Deep Learning Models for Tabular Data
    in their FTTransformer architecture.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset.

        categorical_feature_embedding: ``keras.layers.Layer``,
            An instance of ``CategoricalFeatureEmbedding`` layer to embedd categorical features
            or any layer that can work in its place.
            You can import the ``CategoricalFeatureEmbedding`` layer as follows,
                >>> from teras.layers import CategoricalFeatureEmbedding

        numerical_feature_embedding: ``keras.layers.Layer``,
            An instance of ``FTNumericalFeatureEmbedding`` layer to embedd numerical features
            or any layer that can work in its place.
            You can import the ``FTNumericalFeatureEmbedding`` layer as follows,
                >>> from teras.layers import FTNumericalFeatureEmbedding

        cls_token: ``keras.layers.Layer``,
            An instance of ``FTCLSToken`` layer, it acts as BeRT-like CLS token,
            or any layer than can work in its place.
            You can import the ``FTCLSToken`` layer as follows,
                >>> from teras.layers import FTCLSToken

        encoder: ``keras.layers.Layer``,
            An instance of ``Encoder`` layer to encode the features embeddings,
            or any layer that can work in its palce.
            You can import the `Encoder` layer as follows,
                >>> from teras.layerflow.layers import Encoder

        head: ``keras.layers.Layer``,
            An instance of either ``ClassificationHead`` or ``RegressionHead`` layers,
            depending on the task at hand.

            REMEMBER: In case you're using this model as a base model for pretraining, you MUST leave
            this argument as None.

            You can import the ``ClassificationHead`` and ``RegressionHead`` layers as follows,
                >>> from teras.layers import ClassificationHead
                >>> from teras.layers import RegressionHead
    """
    def __init__(self,
                 input_dim: int,
                 categorical_feature_embedding: keras.layers.Layer = None,
                 numerical_feature_embedding: keras.layers.Layer = None,
                 cls_token: keras.layers.Layer = None,
                 encoder: keras.layers.Layer = None,
                 head: keras.layers.Layer = None,
                 **kwargs):
        if categorical_feature_embedding is None and numerical_feature_embedding is None:
            raise ValueError("Both `categorical_feature_embedding` and `numerical_feature_embedding` "
                             "cannot be None at the same time as a tabular dataset must contains "
                             "features of at least one of the types if not both. ")
        if cls_token is None:
            raise ValueError("`cls_token` cannot be None. Please pass an instance of `FTCLSToken` layer. "
                             "You can import it as, `from teras.layers import FTCLSToken`")

        if encoder is None:
            raise ValueError("`encoder` cannot be None. Please pass an instance of `Encoder` layer. "
                             "You can import it as, `from teras.layerflow.layers import Encoder``")

        if isinstance(input_dim, int):
            input_dim = (input_dim,)

        inputs = keras.layers.Input(shape=input_dim)
        categorical_out = None

        if categorical_feature_embedding is not None:
            x = categorical_feature_embedding(inputs)
            categorical_out = x

        if numerical_feature_embedding is not None:
            x = numerical_feature_embedding(inputs)
            if categorical_out is not None:
                x = keras.layers.Concatenate(axis=1)([categorical_out, x])
        x = cls_token(x)
        outputs = encoder(x)

        if head is not None:
            outputs = head(outputs)

        super().__init__(inputs=inputs,
                         outputs=outputs,
                         **kwargs)
        self.input_dim = input_dim
        self.categorical_feature_embedding = categorical_feature_embedding
        self.numerical_feature_embedding = numerical_feature_embedding
        self.cls_token = cls_token
        self.encoder = encoder
        self.head = head

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'categorical_feature_embedding': keras.layers.serialize(self.categorical_feature_embedding),
                       'numerical_feature_embedding': keras.layers.serialize(self.numerical_feature_embedding),
                       'cls_token': keras.layers.serialize(self.cls_token),
                       'encoder': keras.layers.serialize(self.encoder),
                       'head': keras.layers.serialize(self.head),
                       })
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        categorical_feature_embedding = keras.layers.deserialize(config.pop("categorical_feature_embedding"))
        numerical_feature_embedding = keras.layers.deserialize(config.pop("numerical_feature_embedding"))
        cls_token = keras.layers.deserialize(config.pop("cls_token"))
        encoder = keras.layers.deserialize(config.pop("encoder"))
        head = keras.layers.deserialize(config.pop("head"))
        return cls(input_dim=input_dim,
                   categorical_feature_embedding=categorical_feature_embedding,
                   numerical_feature_embedding=numerical_feature_embedding,
                   cls_token=cls_token,
                   encoder=encoder,
                   head=head,
                   **config)



# Copyright 2023 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# General purpose utility functions
from teras.utils.utils import (tf_random_choice,
                               get_normalization_layer,
                               get_categorical_features_cardinalities,
                               get_activation,
                               get_initializer,
                               get_features_metadata_for_embedding,
                               dataframe_to_tf_dataset,
                               convert_dict_to_array_tensor,
                               serialize_layers_collection,
                               deserialize_layers_collection,
                               inject_missing_values,
                               generate_fake_gemstone_data)


# NODE utility function(s)
from teras.utils.node import sparsemoid


# VIME utility function(s)
from teras.utils.vime import (vime_mask_generator,
                              vime_pretext_generator,
                              preprocess_input_vime_self,
                              preprocess_input_vime_semi)



from teras.layers.node.node_feature_selector import NodeFeatureSelector
import tensorflow as tf


def test_node_features_selector_valid_call_when_max_features_is_none():
    data_dim = 4
    intermediate_inputs = tf.ones((8, 16), dtype=tf.float32)
    feature_selector = NodeFeatureSelector(data_dim=data_dim,
                                           max_features=None)
    outputs = feature_selector(intermediate_inputs)


def test_node_features_selector_valid_call_when_max_features_is_not_none():
    data_dim = 4
    intermediate_inputs = tf.ones((8, 16), dtype=tf.float32)
    feature_selector = NodeFeatureSelector(data_dim=data_dim,
                                           max_features=8)
    outputs = feature_selector(intermediate_inputs)


def test_node_features_selector_output_shape_when_max_features_is_none():
    data_dim = 4
    intermediate_inputs = tf.ones((8, 16), dtype=tf.float32)
    feature_selector = NodeFeatureSelector(data_dim=data_dim,
                                           max_features=None)
    outputs = feature_selector(intermediate_inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    # It must be equal to the dimensions of intermediate dimensions i.e. 16
    assert tf.shape(outputs)[1] == 16


def test_node_features_selector_output_shape_when_max_features_is_not_none():
    data_dim = 4
    intermediate_inputs = tf.ones((8, 16), dtype=tf.float32)
    feature_selector = NodeFeatureSelector(data_dim=data_dim,
                                           max_features=8)
    outputs = feature_selector(intermediate_inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    # It must have data_dim + (min(intermediate_inputs.shape[1], max_features) - data_dimensions) dimensions
    assert tf.shape(outputs)[1] == 8



import tensorflow as tf
import pandas as pd
import numpy as np
from teras.preprocessing.base.base_data_transformer import BaseDataTransformer as _BaseDataTransformer
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from teras.utils.types import FeaturesNamesType
from teras.utils.vime import vime_mask_generator, vime_pretext_generator


class VimeDataTransformer(_BaseDataTransformer):
    """
    DataTransformer class for VIME archictecure.
    It one hot encodes the categorical features
    and normalizes the numerical features.

    Args:
        categorical_features: ``List[str]``,
            List of categorical features names.

        numerical_features: ``List[str]``,
            List of numerical features names.
    """
    def __init__(self,
                 categorical_features: FeaturesNamesType = None,
                 numerical_features: FeaturesNamesType = None
                 ):
        self.categorical_features = categorical_features
        self.numerical_features = numerical_features
        self.one_hot_encoder = OneHotEncoder()

    def fit(self, x: pd.DataFrame):
        if self.categorical_features is not None:
            self.one_hot_encoder.fit(x[self.categorical_features])

    def transform(self, x: pd.DataFrame):
        if self.categorical_features is not None:
            x[self.categorical_features] = self.one_hot_encoder(x[self.categorical_features])

        if self.numerical_features is not None:
            x[self.numerical_features] = np.sum(x[self.numerical_features], axis=1)
        return x


class VimeDataSampler:
    """
    VimeDataSampler class for ``VIME`` architecture.
    It prepares dataset in a format that is acceptable
    for the VIME architecure.

    Args:
        batch_size: ``int``, default 1024,
            Batch size to use.

        shuffle: ``bool``, default True,
            Whether to shuffle the dataset.

        labeled_split: ``float``, default 0.5,
            The size of the original dataset that should be used as labeled
            dataset. ``1 - labeled_split`` will be used as unlabeled dataset for
            in the `VIME` architecture.
            If you have separate unlabeled dataset, pass it along with labeled dataset
            to the  ``get_dataset`` method, in which case this parameter will be ignored
            and your labeled dataset won't be split.

    """
    def __init__(self,
                 batch_size: int = 1024,
                 shuffle: bool = True,
                 labeled_split: float = 0.5,
                 p_m: float = 0.3
                 ):
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.labeled_split = labeled_split
        self.p_m = p_m

    def get_dataset(self,
                    x,
                    y=None,
                    unlabeled_data=None):
        """
        It will return two datasets, one labeled, for semi-supervised training
        and the other unlabeled for self-supervised training.

        Args:
            x:
                Input dataset. It must not contain the target variable.

            y:
                Targets array for ``x`` (the labeled dataset).

            unlabeled_data:
                Unlabeled dataset. If specified, ``x`` won't be split,
                otherwise ``labeled_split`` argument specified during the
                instantiation of the ``VimeDataSampler`` class will be used
                to split ``x``.

        Returns:
            A dataset for self-supervised learning and a dataset for semi-supervised learning.
        """
        labeled_data = x
        if unlabeled_data is None:
            labeled_data, unlabeled_data = train_test_split(x,
                                                            test_size=self.labeled_split,
                                                            shuffle=True)

        # Self supervised dataset
        mask_unlabeled = vime_mask_generator(x=unlabeled_data,
                                             p_m=self.p_m
                                             )
        mask_corrupted, x_tilde = vime_pretext_generator(x=unlabeled_data,
                                                         mask=mask_unlabeled)

        self_ds = tf.data.Dataset.from_tensor_slices((x_tilde,
                                                      {"mask_estimator": mask_corrupted,
                                                       "feature_estimator": mask_unlabeled}))

        if self.batch_size is not None:
            self_ds = self_ds.batch(batch_size=self.batch_size)

        # Semi supervised dataset
        num_samples_labeled = len(labeled_data)
        num_samples_unlabeled = len(unlabeled_data)

        semi_labeled_ds = (tf.data.Dataset
                           .from_tensor_slices({"x_labeled": labeled_data, "y_labeled": y},
                                               name="labeled_dataset")
                           .shuffle(buffer_size=num_samples_labeled,
                                    reshuffle_each_iteration=True)
                           )

        semi_unlabeled_ds = (tf.data.Dataset
                             .from_tensor_slices({"x_unlabeled": unlabeled_data},
                                                 "unlabeled_dataset")
                             .shuffle(buffer_size=num_samples_unlabeled,
                                      reshuffle_each_iteration=True)
                             )

        # If there are fewer unlabeled samples than labeled samples, we'll repeat unlabeled dataset
        # because it is essential to have a batch of unlabeled dataset along with labeled dataset
        # for semi supervised training in VIME.
        # In case when unlabeled dataset contains more samples than labeled, we won't alter anything
        # Because by default, tensorflow dataset will produce batches equal to
        # `math.ceil(min(len(dataset_1), len(dataset_2)) / batch_size)`
        if num_samples_labeled > num_samples_unlabeled:
            num_times_to_repeat = tf.cast(tf.math.ceil(num_samples_labeled / num_samples_unlabeled),
                                          dtype=tf.int64)
            semi_unlabeled_ds = semi_unlabeled_ds.repeat(num_times_to_repeat)

        semi_ds = tf.data.Dataset.zip((semi_labeled_ds, semi_unlabeled_ds),
                                      "semi_supervised_dataset")
        if self.batch_size is not None:
            semi_ds = semi_ds.batch(batch_size=self.batch_size)
        return self_ds, semi_ds



import tensorflow as tf
from tensorflow import keras
from typing import Literal
import math

PERIOD_INITIALIZATIONS = Literal["log-linear", "normal"]


@keras.saving.register_keras_serializable(package="teras.layers.preprocessing")
class PeriodicEmbedding(keras.layers.Layer):
    """
    PeriodicEmbedding layer for numerical features
    as proposed Yury Gorishniy et al. in the paper
    On Embeddings for Numerical Features in Tabular Deep Learning, 2022.

    Reference(s):
        https://arxiv.org/abs/2203.05556

    Args:
        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Dimensionality of numerical embeddings

        initialization: default "normal",
            Initialization strategy.

        sigma: ``float``, default 0.01,
            Used for coefficients initialization
    """
    def __init__(self,
                 features_metadata: dict,
                 embedding_dim: int = 32,
                 initialization: PERIOD_INITIALIZATIONS = "normal",
                 sigma: float = 0.01,
                 **kwargs):
        super().__init__(**kwargs)
        if not len(features_metadata["numerical"]) > 0:
            raise ValueError("`features_metadata` contains no numerical features. "
                             "Either you forgot to pass numerical features names list to the "
                             "`get_features_metadata_for_embedding` or the dataset does not contain "
                             "any numerical features to begin with. \n"
                             "In either case, "
                             "`PeriodicEmbedding` cannot be called on inputs with no numerical features. ")

        if initialization.lower() not in ["normal", "log-linear"]:
            raise ValueError("Invalid value for initialization. Must be one of ['log-linear', 'normal']\n"
                             f"Received: {initialization}")
        self.features_metadata = features_metadata
        self.embedding_dim = embedding_dim
        self.initialization = initialization.lower()
        self.sigma = sigma

        self._numerical_features_indices = list(self.features_metadata["numerical"].values())
        self._num_numerical_features = len(self.features_metadata["numerical"])
        # The official implementation uses another variable n, that is half of the embedding dim
        self.n = self.embedding_dim // 2
        self.coefficients = None

    def build(self, input_shape):
        if self.initialization == "log-linear":
            self.coefficients = self.sigma ** (tf.range(self.n) / self.n)
            self.coefficients = tf.repeat(self.coefficients[None],
                                          repeats=self._num_numerical_features,
                                          axis=1)
        else:
            # initialization must be normal
            self.coefficients = tf.random.normal(shape=(self._num_numerical_features, self.n),
                                                 mean=0.,
                                                 stddev=self.sigma)

        self.coefficients = tf.Variable(self.coefficients)

    @staticmethod
    def cos_sin(x):
        return tf.concat([tf.cos(x), tf.sin(x)], -1)

    def call(self, inputs):
        numerical_features = tf.gather(inputs,
                                       indices=self._numerical_features_indices,
                                       axis=1)
        numerical_features = tf.cast(numerical_features, dtype=tf.float32)

        pi = tf.constant(math.pi)
        return self.cos_sin(2. * pi
                            * tf.expand_dims(self.coefficients, axis=0)
                            * tf.expand_dims(numerical_features, axis=-1))

    def get_config(self):
        config = super().get_config()
        config.update({'features_metadata': self.features_metadata,
                       'embedding_dim': self.embedding_dim,
                       'initialization': self.initialization,
                       'sigma': self.sigma,
                       })
        return config

    @classmethod
    def from_config(cls, config):
        features_metadata = config.pop("features_metadata")
        return cls(features_metadata=features_metadata,
                   **config)



import tensorflow as tf
from teras.layers.node.odst import ObliviousDecisionTree


def test_node_oblivious_decision_tree_valid_call():
    oblivious = ObliviousDecisionTree(num_trees=16,
                                      depth=8,
                                      tree_dim=4)
    inputs = tf.ones((128, 16), dtype=tf.float32)
    outputs = oblivious(inputs)


def test_node_oblivious_decision_tree_output_shape():
    oblivious = ObliviousDecisionTree(num_trees=16,
                                      depth=8,
                                      tree_dim=4)
    inputs = tf.ones((128, 16), dtype=tf.float32)
    outputs = oblivious(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 16 * 4   # num_tress * tree_dim



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.dnfnet")
class DNFNetLocalization(keras.layers.Layer):
    """
    DNFNetLocalization layer based on the localization component
    proposed by Liran Katzir et al.
    in the paper Net-DNF: Effective Deep Modeling of Tabular Data.

    Reference(s):
        https://openreview.net/forum?id=73WTGs96kho

    Args:
        num_formulas: ``int``, default 256,
            Number of DNF formulas to use.
            Each DNF formula is analogous to a tree in tree based ensembles.

        temperature: ``float``, default 2.0,
            Temperature value to use.
            According to the paper, The inclusion of an adaptive temperature in this localization mechanism
            facilitates a data-dependent degree of exclusivity:
            at high temperatures, only a few DNNFs will handle an input instance whereas
            at low temperatures, more DNNFs will effectively participate in the ensemble.
    """
    def __init__(self,
                 num_formulas: int = 256,
                 temperature: float = 2.0,
                 **kwargs):
        super().__init__(**kwargs)
        self.num_formulas = num_formulas
        self.temperature = temperature
        self.softmax = keras.layers.Softmax()
        self.mu = None
        self.sigma = None

    def build(self, input_shape):
        input_dim = input_shape[1]
        mu_initializer = keras.initializers.random_normal()
        self.mu = tf.Variable(initial_value=mu_initializer(shape=(self.num_formulas, input_dim)),
                              shape=(self.num_formulas, input_dim),
                              name='exp_mu')
        sigma_initializer = keras.initializers.random_normal()
        self.sigma = tf.Variable(initial_value=sigma_initializer(shape=(1, self.num_formulas, input_dim)),
                                 shape=(1, self.num_formulas, input_dim),
                                 name="exp_sigma")
        self.temperature = tf.Variable(name='temperature',
                                       initial_value=tf.constant(value=self.temperature),
                                       dtype=tf.float32)

    def call(self, inputs):
        diff = tf.expand_dims(inputs, axis=1) - tf.expand_dims(self.mu, axis=0)
        loc = tf.exp(-1 * tf.norm(tf.multiply(diff, self.sigma), axis=-1))
        loc = self.softmax(tf.sigmoid(self.temperature) * loc)
        return loc

    def get_config(self):
        config = super().get_config()
        config.update({'num_formulas': self.num_formulas,
                       'temperature': self.temperature}
                      )
        return config



from tensorflow import keras
from teras.utils import (serialize_layers_collection,
                         deserialize_layers_collection)
from typing import List, Union

LAYERS_COLLECTION = Union[keras.layers.Layer, List[keras.layers.Layer], keras.models.Model]


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.saint")
class SAINTEncoder(keras.layers.Layer):
    """
    SAINTEncoder layer with LayerFlow desing.
    It is part of the SAINT architecture,
    which is proposed by Gowthami Somepalli et al. in the paper
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.
    It simply stacks N transformer layers and applies them to the outputs
    of the embedded features.

    It differs from the typical Encoder block only in that the Transformer
    layer is a bit different from the regular Transformer layer used in the
    Transformer based architectures as it uses multi-head inter-sample attention,
    in addition to the regular mutli-head attention for features.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        saint_transformer_layers: ``keras.layers.Layer`` or ``List[Layer]`` or ``models.Model``,
            A list, a keras layer or a keras model made up of ``SAINTTransformer`` layers or
            any other custom layers for that purpose.
            You can import the ``SAINTTransformer`` layer as follows,
                >>> from teras.layerflow.layers import SAINTTransformer
    """
    def __init__(self,
                 saint_transformer_layers: LAYERS_COLLECTION,
                 **kwargs):
        super().__init__(**kwargs)
        self.saint_transformer_layers = saint_transformer_layers

        if isinstance(self.saint_transformer_layers, (list, tuple)):
            self.saint_transformer_layers = keras.models.Sequential(self.saint_transformer_layers,
                                                                    name="saint_transformer_layers")

    def call(self, inputs):
        return self.saint_transformer_layers(inputs)

    def get_config(self):
        config = super().get_config()
        config.update({'saint_transformer_layers': serialize_layers_collection(self.saint_transformer_layers)})
        return config

    @classmethod
    def from_config(cls, config):
        saint_transformer_layers = deserialize_layers_collection(config.pop("saint_transformer_layers"))
        return cls(saint_transformer_layers, **config)


import tensorflow as tf
from tensorflow import keras
from teras.losses.ctgan import generator_loss, discriminator_loss
from tqdm import tqdm


@keras.saving.register_keras_serializable(package="keras.layerflow.models")
class CTGAN(keras.Model):
    """
    CTGAN model with LayerFlow design.
    CTGAN is a state-of-the-art tabular data generation architecture
    proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        generator: ``keras.Model``,
            An instance of ``CTGANGenerator`` or any other keras model that can work in
            its place.
            You can import the ``CTGANGenerator`` as follows,
                >>> from teras.models import CTGANGenerator

        discriminator: ``keras.Model``,
            An instance of ``CTGANDiscriminator`` or any other keras model that can work in
            its place.
            You can import the ``CTGANDiscriminator`` as follows,
                >>> from teras.models import CTGANDiscriminator

        num_discriminator_steps: ``int``, default 1,
            Number of discriminator training steps per ``CTGAN`` training step.

        latent_dim: ``int``, default 128,
            Dimensionality of noise or ``z`` that serves as input to ``CTGANGenerator``
            to generate samples.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 num_discriminator_steps: int = 1,
                 latent_dim: int = 128,
                 **kwargs):
        super().__init__(**kwargs)
        self.generator = generator
        self.discriminator = discriminator
        self.num_discriminator_steps = num_discriminator_steps
        self.latent_dim = latent_dim

        # Loss trackers
        self.generator_loss_tracker = keras.metrics.Mean(name="generator_loss")
        self.discriminator_loss_tracker = keras.metrics.Mean(name="discriminator_loss")

    def compile(self,
                generator_optimizer=keras.optimizers.Adam(learning_rate=1e-3,
                                                          beta_1=0.5, beta_2=0.9),
                discriminator_optimizer=keras.optimizers.Adam(learning_rate=1e-3,
                                                              beta_1=0.5, beta_2=0.9),
                generator_loss=generator_loss,
                discriminator_loss=discriminator_loss,
                **kwargs
                ):
        super().compile(**kwargs)
        self.generator_optimizer = generator_optimizer
        self.discriminator_optimizer = discriminator_optimizer
        self.generator_loss = generator_loss
        self.discriminator_loss = discriminator_loss

    def call(self, inputs):
        generated_samples = self.generator(inputs)
        return generated_samples

    def train_step(self, data):
        # real_samples, shuffled_idx, random_features_indices, random_values_indices = data
        real_samples, cond_vectors_real, cond_vectors, mask = data
        self.batch_size = tf.shape(real_samples)[0]

        for _ in range(self.num_discriminator_steps):
            z = tf.random.normal(shape=[self.batch_size, self.latent_dim])
            # cond_vector, mask = self.data_sampler.sample_cond_vector_for_training(self.batch_size,
            #                                                                       random_features_indices=random_features_indices,
            #                                                                       random_values_indices=random_values_indices)
            input_gen = tf.concat([z, cond_vectors], axis=1)
            generated_samples = self.generator(input_gen, training=False)
            # cond_vector_2 = tf.gather(cond_vector, indices=tf.cast(shuffled_idx, tf.int32))
            generated_samples = tf.concat([generated_samples, cond_vectors], axis=1)
            real_samples = tf.concat([real_samples, cond_vectors_real], axis=1)

            with tf.GradientTape(persistent=True) as tape:
                y_generated = self.discriminator(generated_samples)
                y_real = self.discriminator(real_samples)
                grad_pen = self.discriminator.gradient_penalty(real_samples, generated_samples)
                loss_disc = self.discriminator_loss(y_real, y_generated)
            gradients_pen = tape.gradient(grad_pen, self.discriminator.trainable_weights)
            gradients_loss = tape.gradient(loss_disc, self.discriminator.trainable_weights)
            self.discriminator_optimizer.apply_gradients(zip(gradients_pen, self.discriminator.trainable_weights))
            self.discriminator_optimizer.apply_gradients(zip(gradients_loss, self.discriminator.trainable_weights))

        z = tf.random.normal(shape=[self.batch_size, self.latent_dim])
        # cond_vector, mask = self.data_sampler.sample_cond_vector_for_training(self.batch_size,
        #                                                                       random_features_indices=random_features_indices,
        #                                                                       random_values_indices=random_values_indices)
        # Practically speaking, we don't really need the partial function,
        # but it makes things look more neat
        # generator_partial_loss_fn = partial(generator_loss, mask=mask, metadata=self.metadata)
        input_gen = tf.concat([z, cond_vectors], axis=1)
        with tf.GradientTape() as tape:
            tape.watch(cond_vectors)
            tape.watch(mask)
            # generated_samples, y_generated = self(input_gen, cond_vector=cond_vector)
            generated_samples = self(input_gen)
            generated_samples = tf.concat([generated_samples, cond_vectors], axis=1)
            y_generated = self.discriminator(generated_samples, training=False)
            loss_gen = self.generator_loss(generated_samples, y_generated,
                                           cond_vectors=cond_vectors, mask=mask,
                                           metadata=self.metadata)
            # dummy_targets = tf.zeros(shape=(self.batch_size,))
            # loss_gen_dummy = self.generator.compiled_loss(dummy_targets, loss_gen)
        gradients = tape.gradient(loss_gen, self.generator.trainable_weights)
        self.generator_optimizer.apply_gradients(zip(gradients, self.generator.trainable_weights))

        self.generator_loss_tracker.update_state(loss_gen)
        self.discriminator_loss_tracker.update_state(loss_disc)

        results = {m.name: m.result() for m in self.metrics}
        generator_results = {'generator_' + m.name: m.result() for m in self.generator.metrics}
        results.update(generator_results)
        discriminator_results = {'discriminator_' + m.name: m.result() for m in self.discriminator.metrics}
        results.update(discriminator_results)
        return results

    def generate(self,
                 num_samples: int,
                 data_sampler,
                 data_transformer=None,
                 reverse_transform: bool = False,
                 batch_size: int = None):
        """
        Generates new samples using the trained generator.

        Args:
            num_samples: ``int``,
                Number of new samples to generate

            data_sampler:
                Instance of the ``CTGANDataSampler`` class used in preparing
                the tensorflow dataset for training.

            data_transformer:
                Instance of ``CTGANDataTransformer`` class that was used to preprocess the raw data.
                This is required only if the ``reverse_transform`` is set to True.

            reverse_transform: ``bool``, default False,
                Whether to reverse transform the generated data to the original data format.
                If False, the raw generated data will be returned, which you can then manually
                transform into original data format by utilizing ``CTGANDataTransformer`` instance's
                ``reverse_transform`` method.

            batch_size: ``int``, default None.
                If a value is passed, samples will be generated in batches
                where ``batch_size`` determines the size of each batch.
                If ``None``, all ``num_samples`` will be generated at once.
                Note that, if the number of samples to generate aren't huge
                or you know your hardware can handle to generate all samples at once,
                you can leave the value to None, otherwise it is recommended to specify
                a value for batch size.
        """
        if batch_size is None:
            batch_size = num_samples
        num_steps = num_samples // batch_size
        num_steps += 1 if num_samples % batch_size != 0 else 0
        generated_samples = []
        for _ in tqdm(range(num_steps), desc="Generating Data"):
            z = tf.random.normal(shape=[batch_size, self.latent_dim])
            cond_vector = data_sampler.sample_cond_vectors_for_generation(batch_size)
            input_gen = tf.concat([z, cond_vector], axis=1)
            generated_samples.append(self.generator(input_gen, training=False))
        generated_samples = tf.concat(generated_samples, axis=0)
        generated_samples = generated_samples[:num_samples]

        if reverse_transform:
            if data_transformer is None:
                raise ValueError("""To reverse transform the raw generated data, ``data_transformer`` must not be None.
                         Please pass the instance of ``CTGANDataTransformer`` class that was used to transform the
                         input data. Or alternatively, you can set ``reverse_transform`` to False, and later
                         manually reverse transform the generated raw data to original format by utilizing the 
                         ``CTGANDataTransformer`` instance's ``reverse_transform`` method.""")
            generated_samples = data_transformer.reverse_transform(x_generated=generated_samples)

        return generated_samples

    def get_config(self):
        config = super().get_config()
        config.update({'generator': keras.layers.serialize(self.generator),
                       'discriminator': keras.layers.serialize(self.generator),
                       'num_discriminator_steps': self.num_discriminator_steps,
                       'latent_dim': self.latent_dim,
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        generator = keras.layers.deserialize(config.pop("generator"))
        discriminator = keras.layers.deserialize(config.pop("discriminator"))
        return cls(generator=generator,
                   discriminator=discriminator,
                   **config)



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.ft_transformer")
class FTNumericalFeatureEmbedding(keras.layers.Layer):
    """
    Numerical Feature Emebdding layer as proposed by
    Yury Gorishniy et al. in the paper,
    Revisiting Deep Learning Models for Tabular Data
    in their FTTransformer architecture.

    It simply just projects inputs with  dimensions to Embedding dimensions.
    And the only difference between this NumericalFeatureEmbedding
    and the SAINT's NumericalFeatureEmbedding layer that,
    this layer doesn't use a hidden layer. That's it.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Dimensionality of embeddings, must be the same as the one
            used for embedding categorical features.
    """
    def __init__(self,
                 features_metadata: dict,
                 embedding_dim: int = 32,
                 **kwargs
                 ):
        super().__init__(**kwargs)
        if not len(features_metadata["numerical"]) > 0:
            raise ValueError("`features_metadata` contains no numerical features. "
                             "Either you forgot to pass numerical features names list to the "
                             "`get_features_metadata_for_embedding` or the dataset does not contain "
                             "any numerical features to begin with. \n"
                             "In either case, "
                             "`FTNumericalFeatureEmbedding` cannot be called on inputs with no numerical features. ")
        self.features_metadata = features_metadata
        self.numerical_features_metadata = self.features_metadata["numerical"]
        self.embedding_dim = embedding_dim

        self._num_numerical_features = len(self.numerical_features_metadata)
        self._numerical_features_idx = list(self.numerical_features_metadata.values())
        # Need to create as many embedding layers as there are numerical features
        self.embedding_layers = []
        for _ in range(self._num_numerical_features):
            self.embedding_layers.append(
                keras.layers.Dense(units=self.embedding_dim)
            )

    def call(self, inputs):
        numerical_feature_embeddings = tf.TensorArray(size=self._num_numerical_features,
                                                      dtype=tf.float32)

        for i, feature_idx in enumerate(self._numerical_features_idx):
            feature = tf.expand_dims(inputs[:, feature_idx], 1)
            embedding = self.embedding_layers[i]
            feature = embedding(feature)
            numerical_feature_embeddings = numerical_feature_embeddings.write(i, feature)

        numerical_feature_embeddings = tf.squeeze(numerical_feature_embeddings.stack())
        numerical_feature_embeddings = tf.transpose(numerical_feature_embeddings, perm=[1, 0, 2])
        numerical_feature_embeddings.set_shape((None, self._num_numerical_features, self.embedding_dim))
        return numerical_feature_embeddings

    def get_config(self):
        config = super().get_config()
        config.update({'features_metadata': self.features_metadata,
                      'embedding_dim': self.embedding_dim})
        return config

    @classmethod
    def from_config(cls, config):
        features_metadata = config.pop("features_metadata")
        return cls(features_metadata, **config)



import tensorflow as tf
from tensorflow import keras
from teras.layers.tabnet.tabnet_feature_transformer import TabNetFeatureTransformer
from teras.layers.tabnet.tabnet_attentive_transformer import TabNetAttentiveTransformer


@keras.saving.register_keras_serializable("teras.layers.tabnet")
class TabNetEncoder(keras.layers.Layer):
    """
    TabNetEncoder layer based on the TabNet architecture
    proposed by Sercan et al. in the TabNet paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        data_dim: ``int``,
            The dimensionality of the original input dataset,
            or the number of features in the original input dataset.

        feature_transformer_dim: ``int``, default 32,
            The dimensionality of the hidden representation in feature transformation block.
            Each layer first maps the representation to a `2 * feature_transformer_dim`
            output and half of it is used to determine the
            non-linearity of the GLU activation where the other half is used as an
            input to GLU, and eventually `feature_transformer_dim` output is
            transferred to the next layer.

        decision_step_output_dim: ``int``, default 32,
            The dimensionality of output at each decision step, which is later mapped to the
            final classification or regression output.
            It is recommended to keep ``decision_step_output_dim`` and ``feature_transformer_dim``
            equal to each other.
            Adjusting these two parameters values is a good way of obtaining a tradeoff between
            performance and complexity.

        num_decision_steps: ``int``, default 5,
            The number of sequential decision steps.
            For most datasets a value in the range [3, 10] is optimal.
            If there are more informative features in the dataset, the value tends to
            be higher. That said, a very high value of `num_decision_steps` may suffer
            from overfitting.

        num_shared_layers: ``int``, default 2,
            Number of shared layers to use in the ``TabNetFeatureTransformer``.
            These shared layers are *shared* across decision steps.

        num_decision_dependent_layers: ``int``, default 2,
            Number of decision dependent layers to use in the ``TabNetFeatureTransformer``.
            In simple words, ``num_decision_dependent_layers`` are created
            for each decision step in the ``num_decision_steps``.
            For instance, if ``num_decision_steps = `5` and  ``num_decision_dependent_layers = 2``
            then 10 layers will be created, 2 for each decision step.

        relaxation_factor: ``float``, default 1.5,
            Relaxation factor that promotes the reuse of each
            feature at different decision steps. When it is 1, a feature is enforced
            to be used only at one decision step and as it increases, more
            flexibility is provided to use a feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on the performance.
            Typically, a larger value for `num_decision_steps` favors for a larger ``relaxation_factor``.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for ``BatchNormalization`` layer.

        virtual_batch_size: `int`, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.
            This is typically much smaller than the ``batch_size`` used for training.

        residual_normalization_factor: ``float``, default 0.5,
            In the feature transformer, except for the first layer, every other layer utilizes
            normalized residuals, where ``residual_normalization_factor``
            determines the scale of normalization.

        epsilon: ``float``, default 0.00001,
            Epsilon is a small number for numerical stability
            during the computation of entropy loss.
    """
    def __init__(self,
                 data_dim: int,
                 feature_transformer_dim: int = 32,
                 decision_step_output_dim: int = 32,
                 num_decision_steps: int = 5,
                 num_shared_layers: int = 2,
                 num_decision_dependent_layers: int = 2,
                 relaxation_factor: float = 1.5,
                 batch_momentum: float = 0.9,
                 virtual_batch_size: int = 64,
                 residual_normalization_factor: float = 0.5,
                 epsilon=1e-5,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_dim = data_dim
        self.feature_transformer_dim = feature_transformer_dim
        self.decision_step_output_dim = decision_step_output_dim
        self.num_shared_layers = num_shared_layers
        self.num_decision_dependent_layers = num_decision_dependent_layers
        self.num_decision_steps = num_decision_steps
        self.relaxation_factor = relaxation_factor
        self.batch_momentum = batch_momentum
        self.virtual_batch_size = virtual_batch_size
        self.residual_normalization_factor = residual_normalization_factor
        self.epsilon = epsilon

        self.inputs_norm = keras.layers.BatchNormalization(momentum=batch_momentum)

        # To avoid shared layers errors, we should clear them before instantiating
        # the TabNetEncoder layer.
        TabNetFeatureTransformer.reset_shared_layers()
        self.features_transformers_per_step = [TabNetFeatureTransformer(
            units=self.feature_transformer_dim * 2,
            num_shared_layers=self.num_shared_layers,
            num_decision_dependent_layers=self.num_decision_dependent_layers,
            batch_momentum=self.batch_momentum,
            virtual_batch_size=self.virtual_batch_size,
            residual_normalization_factor=self.residual_normalization_factor,
            name=f"step_{i}_feature_transformer"
        )
            for i in range(self.num_decision_steps)
        ]

        # Attentive transformer -- used for creating mask
        self.attentive_transformers_per_step = [TabNetAttentiveTransformer(data_dim=self.data_dim,
                                                                           batch_momentum=self.batch_momentum,
                                                                           virtual_batch_size=self.virtual_batch_size,
                                                                           relaxation_factor=self.relaxation_factor,
                                                                           name=f"step_{i}_attentive_transformer")
                                                for i in range(self.num_decision_steps)
                                                ]
        self.feature_importances_per_sample = []
        self.relu = keras.layers.ReLU()

    def call(self, inputs, mask=None):
        batch_size = tf.shape(inputs)[0]
        # Initializes decision-step dependent variables
        outputs_aggregated = tf.zeros(shape=(batch_size, self.decision_step_output_dim))
        # we only pass mask alongside inputs when we're pretaining
        # As the paper says in the Self-Supervised section, we  initialize
        # mask_values (P) to be 1-S during pretraining.
        # we create this S in the TabNetPretrainer class and pass it from there
        if mask is not None:
            mask_values = mask
        else:
            mask_values = tf.zeros(shape=tf.shape(inputs))
        # Aggregated mask values are used for explaining feature importance
        # Here we'll refer to them as feature_importances_per_sample since
        # that's a more descriptive name
        # For more details, read the `Interpretability` section on page 5
        # of the paper "TabNet: Attentive Interpretable Tabular Learning"
        feature_importances_per_sample = tf.zeros(shape=tf.shape(inputs))
        total_entropy = 0.

        # Prior scales `P` indicate how much a particular feature has been used previously
        # P[i] = ((Pi)j=1 to i)(gamma (gamma - M[j])
        # where `Pi` is the Product Greek letter
        # gamma is the relaxation factor, when gamma = 1, a feature is enforced,
        # to be used only at one decision step and as gamma increases, more
        # flexibility is provided to use a feature at multiple decision steps.
        # P[0] is initialized as all ones, `1`B×D, without any prior
        # on the masked features.
        prior_scales = tf.ones(tf.shape(inputs))

        normalized_inputs = self.inputs_norm(inputs)
        masked_features = normalized_inputs

        for step_num in range(self.num_decision_steps):
            # Feature transformer with two shared and two decision step dependent
            # blocks is used below
            x = self.features_transformers_per_step[step_num](masked_features)

            if step_num > 0 or self.num_decision_steps == 1:
                decision_step_outputs = self.relu(x[:, :self.decision_step_output_dim])
                # Decision aggregation.
                outputs_aggregated += decision_step_outputs
                # Aggregated masks are used for visualization of the
                # feature importance attributes.
                scale_agg = tf.reduce_sum(decision_step_outputs,
                                          axis=1,
                                          keepdims=True)

                # To prevent division by zero, we introduce this conditional
                # and only divide by (num_decision_steps - 1) if num_decision_steps > 1
                if self.num_decision_steps > 1:
                    scale_agg = scale_agg / tf.cast(self.num_decision_steps - 1, tf.float32)

                feature_importances_per_sample += (mask_values * scale_agg)
                self.feature_importances_per_sample.append(feature_importances_per_sample)

            features_for_attentive_transformer = (x[:, self.decision_step_output_dim:])

            if step_num < self.num_decision_steps - 1:
                # Determines the feature masks via linear and nonlinear
                # transformations, taking into account of aggregated feature use
                mask_values = self.attentive_transformers_per_step[step_num](features_for_attentive_transformer,
                                                                             prior_scales=prior_scales)
                # Relaxation factor controls the amount of reuse of features between
                # different decision blocks and updated with the values of coefficients
                prior_scales *= (self.relaxation_factor - mask_values)

                # Entropy is used to penalize the amount of sparsity in feature selection
                total_entropy += tf.reduce_mean(
                    tf.reduce_sum(-mask_values * tf.math.log(mask_values + self.epsilon), axis=1)
                ) / (self.num_decision_steps - 1)
                entropy_loss = total_entropy

                # Feature Selection
                masked_features = tf.multiply(mask_values, normalized_inputs)
            else:
                entropy_loss = 0.

        # Add entropy loss
        self.add_loss(entropy_loss)

        return outputs_aggregated

    def get_config(self):
        config = super().get_config()
        new_config = {'data_dim': self.data_dim,
                      'feature_transformer_dim': self.feature_transformer_dim,
                      'decision_step_output_dim': self.decision_step_output_dim,
                      'num_decision_steps': self.num_decision_steps,
                      'num_shared_layers': self.num_shared_layers,
                      'num_decision_dependent_layers': self.num_decision_dependent_layers,
                      'relaxation_factor': self.relaxation_factor,
                      'batch_momentum': self.batch_momentum,
                      'virtual_batch_size': self.virtual_batch_size,
                      'residual_normalization_factor': self.residual_normalization_factor,
                      'epsilon': self.epsilon,
                      }
        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim, **config)



from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layerflow.layers.saint")
class SAINTTransformer(keras.layers.Layer):
    """
    SAINT Transformer layer with LayerFlow design.
    It is part of the SAINT architecture,
    which is proposed by Gowthami Somepalli et al.
    in the paper SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.
    It differs from the usual Transformer (L) block in that it contains additional
    ``MultiHeadInterSampleAttention`` layer in addition to the usual
    ``MultiHeadAttention`` layer.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        multi_head_inter_sample_attention: ``keras.layers.Layer``,
            An instance of ``MultiHeadInterSampleAttention`` layer or any other custom
            layer that can work in its place.
            You can import this layer as follows,
                >>> from teras.layers import MultiHeadInterSampleAttention

        feed_forward: ``keras.layers.Layer``,
            An instance of ``FeedForward`` layer or any custom layer that can work
            in its place.
            You can import this layer as follows,
                >>> from teras.layerflow.layers import FeedForward

        transformer: ``keras.layers.Layer``,
            An instance of the regular ``Transformer`` layer, or any custom layer
            that can work in its place.
            You can import this layer as follows,
                >>> from teras.layerflow.layers import Transformer
    """
    def __init__(self,
                 multi_head_inter_sample_attention: keras.layers.Layer,
                 feed_forward: keras.layers.Layer,
                 transformer: keras.layers.Layer,
                 **kwargs):
        super().__init__(**kwargs)
        self.multi_head_inter_sample_attention = multi_head_inter_sample_attention
        self.feed_forward = feed_forward
        self.transformer = transformer

    def build(self, input_shape):
        # We build the inner SAINT Transformer block using keras Functional API
        # and since we need the input dimensions that's why we're building it in the build method.

        # Inter Sample Attention Block: this attention is applied to rows.
        inputs = keras.layers.Input(shape=tuple(input_shape[1:]))
        intermediate_outputs = inputs
        if self.apply_attention_to_rows:
            residual = inputs
            x = self.multi_head_inter_sample_attention(inputs)
            x = keras.layers.Add()([x, residual])
            x = keras.layers.LayerNormalization()(x)
            residual = x
            x = self.feed_forward(x)
            x = keras.layers.Add()([x, residual])
            intermediate_outputs = keras.layers.LayerNormalization()(x)
            final_outputs = intermediate_outputs

        # MultiHeadAttention block: this attention is applied to columns
        if self.apply_attention_to_features:
            # If `apply_attention_to_features` is set to True,
            # then attention will be applied to columns/features
            # The MultiHeadInterSampleAttention appollection)lies attention over rows,
            # but the regular MultiHeadAttention layer is used to apply attention over features.
            # Since the common Transformer layer applies MutliHeadAttention over features
            # as well as takes care of applying all the preceding and following layers,
            # so we'll just use that here.
            final_outputs = self.transformer(intermediate_outputs)

        self.transformer_block = keras.Model(inputs=inputs,
                                             outputs=final_outputs,
                                             name="saint_inner_transformer_block")

    def call(self, inputs):
        outputs = self.transformer_block(inputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        new_config = {'multi_head_inter_sample_attention': keras.layers.serialize(self.multi_head_inter_sample_attention),
                      'feed_forward': keras.layers.serialize(self.feed_forward),
                      'transformer': keras.layers.serialize(self.transformer),
                      }
        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        multi_head_inter_sample_attention = keras.layers.deserialize(config.pop("multi_head_inter_sample_attention"))
        feed_forward = keras.layers.deserialize(config.pop("feed_forward"))
        transformer = keras.layers.deserialize(config.pop("transformer"))
        return cls(multi_head_inter_sample_attention=multi_head_inter_sample_attention,
                   feed_forward=feed_forward,
                   transformer=transformer,
                   **config)



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.tabtransformer")
class TabTransformerColumnEmbedding(keras.layers.Layer):
    """
    Column Embedding layer as proposed by Xin Huang et al. in the paper
    TabTransformer: Tabular Data Modeling Using Contextual Embeddings.

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        num_categorical_features: ``int``,
            Number of categorical features in the dataset.

        embedding_dim: ``int``, default 32,
            Dimensionality of the embedded features
    """
    def __init__(self,
                 num_categorical_features: int,
                 embedding_dim=32,
                 **kwargs):
        super().__init__(**kwargs)
        self.embedding_dim = embedding_dim
        self.num_categorical_features = num_categorical_features
        self.column_embedding = keras.layers.Embedding(input_dim=self.num_categorical_features,
                                                       output_dim=self.embedding_dim)
        self.column_indices = tf.range(start=0,
                                       limit=self.num_categorical_features,
                                       delta=1)
        self.column_indices = tf.cast(self.column_indices, dtype="float32")

    def call(self, inputs):
        """
        Args:
            inputs: Embeddings of categorical features encoded by CategoricalFeatureEmbedding layer
        """
        return inputs + self.column_embedding(self.column_indices)

    def get_config(self):
        config = super().get_config()
        config.update({'num_categorical_features': self.num_categorical_features,
                      'embedding_dim': self.embedding_dim,
                      })
        return config

    @classmethod
    def from_config(cls, config):
        num_categorical_features = config.pop("num_categorical_features")
        return cls(num_categorical_features, **config)




import tensorflow as tf
from tensorflow import keras
from teras.layers.ctgan.ctgan_generator_block import CTGANGeneratorBlock
from teras.layers.ctgan.ctgan_discriminator_block import CTGANDiscriminatorBlock
from teras.layerflow.models.ctgan import CTGAN as _CTGAN_LF
from teras.layers.activation import GumbelSoftmax
from teras.utils.types import UnitsValuesType


@keras.saving.register_keras_serializable(package="teras.models")
class CTGANGenerator(keras.Model):
    """
    CTGANGenerator for CTGAN architecture as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        data_dim: ``int``,
            The dimensionality of the dataset.
            It will also be the dimensionality of the output produced
            by the generator.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionailty of the transformed dataset throught the
            ``.data_dim`` attribute of the ``CTGANDataSampler`` instance used in sampling
            the dataset.

        metadata: ``dict``,
            The ``CTGANGeneratorBlock``, applies different activation functions
            to its outputs depending on the type of features (categorical or numerical).
            And to determine the feature types and for other computations during the
            activation step, the ``metadata`` computed during the data transformation step,
            is required.

            It can be accessed through the ``.get_metadata()`` method of the ``CTGANDataTransformer``
            instance which was used to transform the raw input data.

            Note that, this is NOT the same metadata as ``features_metadata``, which is computed
            using the ``get_features_metadata_for_embedding`` utility function from ``teras.utils``.
            You must use the ``.get_metadata()`` method of the ``CTGANDataTransformer`` to access it.

            You can import the ``CTGANDataTransformer`` as follows,
                >>> from teras.preprocessing import CTGANDataTransformer

        units_values: ``List[int]`` or ``Tuple[int]``, default [256, 256],
            A list or tuple of units.
            For each value, a ```CTGANGeneratorBlock``` of that
            dimensionality (units) is added to the ``CTGANGeneratorBlock``
            to form its ``hidden block``.
            You can access the ``CTGANGeneratorBlock`` as follows,
                >>> from teras.layers import CTGANGeneratorBlock
    """
    def __init__(self,
                 data_dim: int,
                 metadata: dict = None,
                 units_values: UnitsValuesType = (256, 256),
                 **kwargs):
        super().__init__(**kwargs)

        if not isinstance(units_values, (list, tuple)):
            raise ValueError(f"""`units_values` must be a list or tuple of units which determines
                        the number of Generator residual blocks and the dimensionality of those blocks.
                        Received: {type(units_values)}""")

        if metadata is None:
            raise ValueError(f"""`metadata` cannot be None.
                `metadata`, which is computed during the data transformation step,
                is required by the `CTGANGenerator` to apply relevant activation functions to the 
                output of the Generator. But received `None`.\n
                Please pass the metadata by accessing it through the `.get_metadata()` method
                of the `CTGANDataTransformer` instance which was used to transform the raw input data.
                """)
        self.data_dim = data_dim
        self.metadata = metadata
        self.units_values = units_values

        self.hidden_block = keras.models.Sequential(name="generator_hidden_block")
        for units in self.units_values:
            self.hidden_block.add(CTGANGeneratorBlock(units))
        self.output_layer = keras.layers.Dense(self.data_dim, name="generator_output_layer")
        self.gumbel_softmax = GumbelSoftmax()

    def apply_activations_by_feature_type(self, interim_outputs):
        """
        This function applies activation functions to the interim outputs of
        the Generator by feature type.
        As CTGAN architecture requires specific transformations on the raw input data,
        that decompose one feature in several features,
        and since each type of feature, i.e. numerical or categorical require
        different activation functions to be applied, the process of applying those
        activations becomes rather tricky as it requires knowledge of underlying
        data transformation and features metadata.
        To ease the user's burden, in case a user wants to subclass this
        Generator model and completely customize the inner workings of the generator
        but would want to use the activation method specific to the CTGAN architecture,
        so that the subclassed Generator can work seamlessly with the rest of the
        architecture and there won't be any discrepancies in outputs produced by
        the subclasses Generator and those expected by the architecture,
        this function is separated, so user can just call this function on the interim
        outputs in the `call` method.

        Args:
            interim_outputs: Outputs produced by the `output_layer` of the Generator.

        Returns:
            Final outputs activated by the relevant activation functions.
        """
        outputs = []
        numerical_features_relative_indices = self.metadata["numerical"]["relative_indices_all"]
        features_relative_indices_all = self.metadata["relative_indices_all"]
        num_valid_clusters_all = self.metadata["numerical"]["num_valid_clusters_all"]
        cont_i = 0
        cat_i = 0
        num_categories_all = self.metadata["categorical"]["num_categories_all"]
        for i, index in enumerate(features_relative_indices_all):
            # the first k = num_numerical_features are numerical in the data
            if i < len(numerical_features_relative_indices):
                # each numerical features has been transformed into num_valid_clusters + 1 features
                # where the first feature is alpha while the following features are beta components
                alphas = tf.nn.tanh(interim_outputs[:, index])
                alphas = tf.expand_dims(alphas, 1)
                outputs.append(alphas)
                betas = self.gumbel_softmax(interim_outputs[:, index + 1: index + 1 + num_valid_clusters_all[cont_i]])
                outputs.append(betas)
                cont_i += 1
            # elif index in categorical_features_relative_indices:
            else:
                # each categorical feature has been converted into a one hot vector
                # of size num_categories
                ds = self.gumbel_softmax(interim_outputs[:, index: index + num_categories_all[cat_i]])
                outputs.append(ds)
                cat_i += 1
        outputs = tf.concat(outputs, axis=1)
        return outputs

    def call(self, inputs):
        # inputs have the shape |z| + |cond|
        # while the outputs will have the shape of equal to (batch_size, transformed_data_dims)
        interim_outputs = self.output_layer(self.hidden_block(inputs))
        outputs = self.apply_activations_by_feature_type(interim_outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({'data_dim': self.data_dim,
                       'units_values': self.units_values,
                       'metadata': self.metadata}
                      )
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim,
                   **config)


@keras.saving.register_keras_serializable(package="teras.models")
class CTGANDiscriminator(keras.Model):
    """
    CTGANDiscriminator for CTGAN architecture as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        units_values: ``List[int]`` or ``Tuple[int]``, default [256, 256],
            A list or tuple of units.
            For each value, a ```CTGANDiscriminatorBlock```
            of that dimensionality (units) is added to the discriminator
            to form the `hidden block` of the discriminator.

            You can import the ``CTGANDiscriminatorBlock`` as follows,
                >>> from teras.layers import CTGANDiscriminatorBlock

        packing_degree: ``int``, default 8,
            Packing degree - taken from the PacGAN paper.
            The number of samples concatenated or "packed" together.
            It must be a factor of the batch_size.
            Packing degree is borrowed from the PacGAN [Diederik P. Kingma et al.] architecture,
            which proposes passing `m` samples at once to discriminator instead of `1` to be
            jointly classified as real or fake by the discriminator in order to tackle the
            issue of mode collapse inherent in the GAN based architectures.
            The number of samples passed jointly `m`, is termed as the `packing degree`.

        gradient_penalty_lambda: ``float``, default 10,
                Controls the strength of gradient penalty.
                lambda value is directly proportional to the strength of gradient penalty.
                Gradient penalty penalizes the discriminator for large weights in an attempt
                to combat Discriminator becoming too confident and overfitting.

    Example:
        ```python
        # Instantiate Generator
        generator = CTGANGenerator(data_dim=data_dim,
                                   metadata=metadata)

        # Instantiate Discriminator
        discriminator = CTGANDiscriminator()

        # Sample noise to generate samples from
        z = tf.random.normal([512, 18])

        # Generate samples
        generated_samples = generator(z)

        # Predict using discriminator
        y_pred = discriminator(generated_samples)
        ```
    """
    def __init__(self,
                 units_values: UnitsValuesType = (256, 256),
                 packing_degree: int = 8,
                 gradient_penalty_lambda: float = 10,
                 **kwargs):
        super().__init__(**kwargs)

        if not isinstance(units_values, (list, tuple)):
            raise ValueError(f"""`units_values` must be a list or tuple of units which determines
                        the number of Discriminator blocks and the dimensionality of those blocks.
                        But {units_values} was passed""")

        self.units_values = units_values
        self.packing_degree = packing_degree
        self.gradient_penalty_lambda = gradient_penalty_lambda

        self.hidden_block = keras.models.Sequential(name="discriminator_hidden_block")
        for units in self.units_values:
            self.hidden_block.add(CTGANDiscriminatorBlock(units))

        self.output_layer = keras.layers.Dense(1, name="discriminator_output_layer")

    @tf.function
    def gradient_penalty(self,
                         real_samples,
                         generated_samples):
        """
        Calculates the gradient penalty as proposed
        in the paper "Improved Training of Wasserstein GANs"

        Reference(s):
            https://arxiv.org/abs/1704.00028

        Args:
            real_samples: Data samples drawn from the real dataset
            generated_samples: Data samples generated by the generator

        Returns:
            Gradient penalty computed for given values.
        """
        batch_size = tf.shape(real_samples)[0]
        dim = tf.shape(real_samples)[1]

        alpha = tf.random.uniform(shape=(batch_size // self.packing_degree, 1, 1))
        alpha = tf.reshape(tf.tile(alpha, [1, self.packing_degree, dim]),
                           (-1, dim))
        interpolated_samples = (alpha * real_samples) + ((1 - alpha) * generated_samples)
        with tf.GradientTape() as tape:
            tape.watch(interpolated_samples)
            y_interpolated = self(interpolated_samples)
        gradients = tape.gradient(y_interpolated, interpolated_samples)
        gradients = tf.reshape(gradients, shape=(-1, self.packing_degree * dim))

        # Calculating gradient penalty
        gradients_norm = tf.norm(gradients)
        gradient_penalty = tf.reduce_mean(tf.square(gradients_norm - 1.0)) * self.gradient_penalty_lambda
        return gradient_penalty

    def call(self, inputs):
        inputs_dim = tf.shape(inputs)[1]
        inputs = tf.reshape(inputs, shape=(-1, self.packing_degree * inputs_dim))
        outputs = self.hidden_block(inputs)
        outputs = self.output_layer(outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        new_config = {'units_values': self.units_values,
                      'packing_degree': self.packing_degree,
                      'gradient_penalty_lambda': self.gradient_penalty_lambda}
        config.update(new_config)
        return config


@keras.saving.register_keras_serializable(package="teras.models")
class CTGAN(_CTGAN_LF):
    """
    CTGAN is a state-of-the-art tabular data generation architecture
    proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        input_dim: ``int``,
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionailty of the transformed dataset throught the
            ``.data_dim`` attribute of the ``CTGANDataSampler`` instance used in sampling
            the dataset.

        metadata: ``dict``,
            A dictionary of metadata computed during the data transformation step.
            Simply pass the result of ``.get_metadata()`` method of the ``CTGANDataTransformer`` instance
            which was used to transform the raw input data.
            The Generator in ``CTGAN`` architecture applies different activation functions
            to the output of Generator, depending on the type of features.
            And to determine the feature types and for other computation during
            activation step, the ``metadata`` computed during the data transformation step,
            is required.
            It is also required during the computation of generator loss.
            Hence, it cannot be None.

        generator_units_values: ``List[int]`` or ``Tuple[int]``, default [256, 256],
            A list or tuple of units.
            For each value, a ``CTGANGeneratorBlock``
            of that dimensionality (units) is added to the generator
            to form the ``hidden block`` of the generator.
            You can import the ``CTGANGeneratorBlock`` as follows,
                >>> from teras.layers import CTGANGeneratorBlock

        discriminator_units_values: ``List[int]`` or ``Tuple[int]``, default [256, 256],
            A list or tuple of units values.
            For each value, a ``CTGANDiscriminatorBlock``
            of that dimensionality (units) is added to the discriminator
            to form the ``hidden block`` of the discriminator.
            You can import the ``CTGANDiscriminatorBlock`` as follows,
                >>> from teras.layers import CTGANDiscriminatorBlock

        num_discriminator_steps: ``int``, default 1,
            Number of discriminator training steps per ``CTGAN`` training step.

        latent_dim: ``int``, default 128,
            Dimensionality of noise or `z` that serves as input to Generator
            to generate samples.

        packing_degree: ``int``, default 8,
            Packing degree - taken from the PacGAN paper.
            The number of samples concatenated or "packed" together.
            It must be a factor of the batch_size.
            Packing degree is borrowed from the PacGAN [Diederik P. Kingma et al.] architecture,
            which proposes passing ``m`` samples at once to discriminator instead of ``1`` to be
            jointly classified as real or fake by the discriminator in order to tackle the
            issue of mode collapse inherent in the GAN based architectures.
            The number of samples passed jointly ``m``, is termed as the ``packing degree``.

        gradient_penalty_lambda: ``float``, default 10,
            Controls the strength of gradient penalty in the ``CTGANDiscriminator``.
            lambda value is directly proportional to the strength of gradient penalty.
            Gradient penalty penalizes the discriminator for large weights in an attempt
            to combat Discriminator becoming too confident and overfitting.
    """
    def __init__(self,
                 input_dim: int,
                 metadata: dict = None,
                 generator_units_values: UnitsValuesType = (256, 256),
                 discriminator_units_values: UnitsValuesType = (256, 256),
                 num_discriminator_steps: int = 1,
                 latent_dim: int = 128,
                 packing_degree: int = 8,
                 gradient_penalty_lambda: float = 10,
                 **kwargs):
        if metadata is None:
            raise ValueError("`metadata` cannot be None. "
                             "You can access the `metadata` through `.get_metadata()` method of `CTGANDataTransformer` "
                             "instance.")
        generator = CTGANGenerator(data_dim=input_dim,
                                   metadata=metadata,
                                   units_values=generator_units_values)

        discriminator = CTGANDiscriminator(units_values=discriminator_units_values,
                                           packing_degree=packing_degree,
                                           gradient_penalty_lambda=gradient_penalty_lambda)
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         num_discriminator_steps=num_discriminator_steps,
                         latent_dim=latent_dim,
                         **kwargs)
        self.input_dim = input_dim
        self.metadata = metadata
        self.generator_units_values = generator_units_values
        self.discriminator_units_values = discriminator_units_values
        self.num_discriminator_steps = num_discriminator_steps
        self.latent_dim = latent_dim
        self.packing_degree = packing_degree
        self.gradient_penalty_lambda = gradient_penalty_lambda

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'metadata': self.metadata,
                  'generator_units_values': self.generator_units_values,
                  'discriminator_units_values': self.discriminator_units_values,
                  'num_discriminator_steps': self.num_discriminator_steps,
                  'latent_dim': self.latent_dim,
                  'packing_degree': self.packing_degree,
                  'gradient_penalty_lambda': self.gradient_penalty_lambda,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim=input_dim,
                   **config)



from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.gain")
class GAINDiscriminatorBlock(keras.layers.Layer):
    """
    GAINDiscriminatorBlock that is used in construction of hidden block
    inside the ``GAINDiscriminator`` model of ``GAIN`` architecture.

    Args:
        units: ``int``,
            Hidden dimensionality of the discriminator block

        activation: default "relu",
            Activation function to use in the discriminator block

        kernel_initializer: default "glorot_normal",
            Kernel initializer for the discriminator block
    """
    def __init__(self,
                 units: int,
                 activation="relu",
                 kernel_initializer="glorot_normal",
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.kernel_initializer = kernel_initializer
        self.dense = keras.layers.Dense(units,
                                        activation=self.activation,
                                        kernel_initializer=self.kernel_initializer)

    def call(self, inputs):
        return self.dense(inputs)

    def get_config(self):
        config = super().get_config()
        new_config = {'units': self.units,
                      'activation': self.activation,
                      'kernel_initializer': self.kernel_initializer,
                      }

        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        units = config.pop("units")
        return cls(units, **config)


import tensorflow as tf
from tensorflow import keras
from teras.activations import glu


@keras.saving.register_keras_serializable("teras.layers.tabnet")
class TabNetFeatureTransformerBlock(keras.layers.Layer):
    """
    Feature Transformer block layer is used in constructing the FeatureTransformer
    as proposed by Sercan et al. in TabNet paper.
    It applies a Dense layer followed by a BatchNormalization layer
    followed by a glu activation function.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        units: ``int``, default 32,
            the dimensionality of the hidden representation in feature transformation block.
            Each layer first maps the representation to a ``2 * feature_transformer_dim``
            output and half of it is used to determine the
            non-linearity of the ``glu`` activation where the other half is used as an
            input to ``glu``, and eventually ``feature_transformer_dim`` output is
            transferred to the next layer.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for ``BatchNormalization`` layer.

        virtual_batch_size: ``int``, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.
            This is typically much smaller than the ``batch_size`` used for training.
            And most importantly, the ``batch_size`` must always be divisible by the ``virtual_batch_size``.

        residual_normalization_factor: ``float``, default 0.5,
            In the feature transformer, except for the layer, every other layer utilizes normalized residuals,
            where ``residual_normalization_factor`` determines the scale of normalization.

        use_residual_normalization: ``bool``, default True,
            Whether to use residual normalization.
            According to the default architecture, every layer uses residual normalization
            EXCEPT for the very first layer.
    """
    def __init__(self,
                 units: int = 32,
                 batch_momentum: float = 0.9,
                 virtual_batch_size: int = 64,
                 residual_normalization_factor: float = 0.5,
                 use_residual_normalization: bool = True,
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.batch_momentum = batch_momentum
        self.virtual_batch_size = virtual_batch_size
        self.use_residual_normalization = use_residual_normalization
        self.residual_normalization_factor = residual_normalization_factor
        self.dense = keras.layers.Dense(self.units * 2, use_bias=False)
        self.norm = keras.layers.BatchNormalization(momentum=self.batch_momentum,
                                                    virtual_batch_size=virtual_batch_size)
        self.add = keras.layers.Add()

    def call(self, inputs):
        x = self.dense(inputs)
        x = self.norm(x)
        x = glu(x)
        if self.use_residual_normalization:
            x = self.add([x, inputs]) * tf.math.sqrt(self.residual_normalization_factor)
        return x

    def get_config(self):
        config = super().get_config()
        config.update({'units': self.units,
                       'batch_momentum': self.batch_momentum,
                       'virtual_batch_size': self.virtual_batch_size,
                       'residual_normalization_factor': self.residual_normalization_factor,
                       'use_residual_normalization': self.use_residual_normalization,
                       })
        return config



from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.ctgan")
class CTGANDiscriminatorBlock(keras.layers.Layer):
    """
    Discriminator Block based on the architecture proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    outputs = Dropout(LeakyReLU(Dense(inputs)))

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        units: ``int``, default 256,
            Dimensionality of the hidden layer

        leaky_relu_alpha: ``float``, default 0.2,
            Alpha value to use for leaky relu activation

        dropout_rate: ``float``, default 0.,
            Dropout rate to use in the ``Dropout`` layer,
            which is applied after hidden layer.
    """
    def __init__(self,
                 units: int = 256,
                 leaky_relu_alpha: float = 0.2,
                 dropout_rate: float = 0.,
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.leaky_relu_alpha = leaky_relu_alpha
        self.dropout_rate = dropout_rate

        self.dense = keras.layers.Dense(units)
        self.leaky_relu = keras.layers.LeakyReLU(alpha=self.leaky_relu_alpha)
        self.dropout = keras.layers.Dropout(rate=self.dropout_rate)

    def call(self, inputs):
        out = self.dense(inputs)
        out = self.leaky_relu(out)
        out = self.dropout(out)
        return out

    def get_config(self):
        config = super().get_config()
        config.update({'units': self.units,
                       'leaky_relu_alpha': self.leaky_relu_alpha,
                       'dropout_rate': self.dropout_rate}
                      )
        return config




import tensorflow as tf


def reconstruction_loss(real_samples=None,
                        reconstructed_samples=None,
                        mask=None):
    """
    Reconstruction loss for TabNet Pretrainer mode as propsoed by
     Sercan et al. in the paper,
     "TabNet: Attentive Interpretable Tabular Learning"

     Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        real_samples: Samples drawn from the input dataset
        reconstructed_samples: Samples reconstructed by the decoder
        mask: Mask that indicates the missing of features in a sample

    Returns:
        Reconstruction loss for TabNet Pretraining.
    """
    nominator_part = (reconstructed_samples - real_samples) * mask
    real_samples_population_std = tf.math.reduce_std(tf.cast(real_samples, dtype=tf.float32))
    # divide
    x = nominator_part / real_samples_population_std
    # Calculate L2 norm
    loss = tf.sqrt(tf.reduce_sum(tf.square(x)))
    return loss


import tensorflow as tf
from teras.layers.dnfnet.dnfnet_localization import DNFNetLocalization


def test_dnfnet_localitzation_valid_call():
    localization = DNFNetLocalization(num_formulas=16)
    inputs = tf.ones((8, 10), dtype=tf.float32)
    outputs = localization(inputs)


def test_dnfnet_localitzation_output_shape():
    localization = DNFNetLocalization(num_formulas=16)
    inputs = tf.ones((8, 10), dtype=tf.float32)
    outputs = localization(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 16



# Copyright 2023 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# NODE models
from teras.models.node import (NODE,
                               NODEClassifier,
                               NODERegressor)

# TabNet models
from teras.models.tabnet import (TabNet,
                                 TabNetClassifier,
                                 TabNetRegressor,
                                 TabNetPretrainer)

# TabTransformer models
from teras.models.tabtransformer import (TabTransformer,
                                         TabTransformerClassifier,
                                         TabTransformerRegressor)


# DNFNet models
from teras.models.dnfnet import (DNFNet,
                                 DNFNetRegressor,
                                 DNFNetClassifier)

# SAINT models
from teras.models.saint import (SAINT,
                                SAINTClassifier,
                                SAINTRegressor)


# RTDL ResNet models
from teras.models.rtdl_resnet import (RTDLResNetClassifier,
                                      RTDLResNetRegressor)

# RTDL FTTransformer models
from teras.models.ft_transformer import (FTTransformer,
                                         FTTransformerClassifier,
                                         FTTransformerRegressor)

# VIME models
from teras.models.vime import (VimeSelf,
                               VimeSemi)


# CTGAN models
from teras.models.ctgan import (CTGANGenerator,
                                CTGANDiscriminator,
                                CTGAN)


# TVAE models
from teras.models.tvae import (TVAEEncoder,
                               TVAEDecoder,
                               TVAE)


# GAIN models
from teras.models.gain import (GAINGenerator,
                               GAINDiscriminator,
                               GAIN)


# PCGAIN models
from teras.models.pcgain import (PCGAINGenerator,
                                 PCGAINDiscriminator,
                                 PCGAINClassifier,
                                 PCGAIN)



import tensorflow as tf
from tensorflow import keras
import tensorflow_probability as tfp
from teras.losses.gain import generator_loss, discriminator_loss


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class GAIN(keras.Model):
    """
    GAIN model with LayerFlow design.
    GAIN is a missing data imputation model based on GANs.
    This is an implementation of the GAIN architecture
    proposed by Jinsung Yoon et al. in the paper
    GAIN: Missing Data Imputation using Generative Adversarial Nets.

    In GAIN, the generator observes some components of
    a real data vector, imputes the missing components
    conditioned on what is actually observed, and
    outputs a completed vector.
    The discriminator then takes a completed vector
    and attempts to determine which components
    were actually observed and which were imputed.
    It also utilizes a novel hint mechanism, which
    ensures that generator does in fact learn to generate
    samples according to the true data distribution.

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        generator: ``keras.Model``,
            An instance of ``GAINGenerator`` model or any customized model that can
            work in its place.
            This allows you to take full control over the Generator's architecture.
            You import the standalone ``GAINGenerator`` model as follows,
                >>> from teras.models import GAINGenerator

        discriminator: ``keras.Model``,
            An instance of ``GAINDiscriminator`` model or any customized model that
            can work in its place.
            This allows you to take full control over the Discriminator's architecture.
            You import the standalone ``GAINDiscriminator`` model as follows,
                >>> from teras.models import GAINDiscriminator

        num_discriminator_steps: ``int``, default 1,
            Number of discriminator training steps per GAIN training step.

        hint_rate: ``float``, default 0.9,
            Hint rate will be used to sample binary vectors for
            `hint vectors` generation. Must be between 0. and 1.
            Hint vectors ensure that generated samples follow the
            underlying data distribution.

        alpha: ``float``, default 100,
            Hyper parameter for the generator loss computation that
            controls how much weight should be given to the MSE loss.
            Precisely, `generator_loss` = `cross_entropy_loss` + `alpha` * `mse_loss`
            The higher the `alpha`, the more the mse_loss will affect the
            overall generator loss.

    Example:
        ```python
        input_data = pd.DataFrame(tf.random.uniform([200, 5]),
                                columns=["A", "B", "C", "D", "E"])

        from teras.utils.gain import inject_missing_values
        input_data = inject_missing_values(input_data)

        # GAIN requires data in a specific format for which we have
        # relevant DataSampler and DataTrasnformer classes
        from teras.preprocessing.gain import DataSampler, DataTransformer

        numerical_features = ["A", "B"]
        categorical_features = ["C", "D", "E"]

        data_transformer = DataTransformer(numerical_features=numerical_features,
                                           categorical_features=categorical_features)

        transformed_data = data_transformer.transform(input_data, return_dataframe=True)

        data_sampler = DataSampler()
        dataset = data_sampler.get_dataset(transformed_data)

        # Instantiate GAIN
        gain_imputer = GAIN()

        # Compile it
        gain_imputer.compile()

        # Train it
        gain_imputer.fit(dataset)

        # Predict
        test_data = transformed_data[:50]

        # Imputation Method 1:
        imputed_data = gain_imputer.predict(test_data)

        # Reverse transform into original format
        imputed_data = data_transformer.reverse_transform(imputed_data)

        # Imputation Method 2:
        imputed_data = gain_imputer.impute(test_data)
        ```
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 num_discriminator_steps: int = 1,
                 hint_rate: float = 0.9,
                 alpha: float = 100,
                 **kwargs):
        super().__init__(**kwargs)
        self.generator = generator
        self.discriminator = discriminator
        self.num_discriminator_steps = num_discriminator_steps
        self.hint_rate = hint_rate
        self.alpha = alpha

        self.z_sampler = tfp.distributions.Uniform(low=0.,
                                                   high=0.01,
                                                   name="z_sampler")
        self.hint_vectors_sampler = tfp.distributions.Binomial(total_count=1,
                                                               probs=self.hint_rate,
                                                               name="hint_vectors_generator")

        # Loss trackers
        self.generator_loss_tracker = keras.metrics.Mean(name="generator_loss")
        self.discriminator_loss_tracker = keras.metrics.Mean(name="discriminator_loss")

    def compile(self,
                generator_optimizer=keras.optimizers.Adam(),
                discriminator_optimizer=keras.optimizers.Adam(),
                generator_loss=generator_loss,
                discriminator_loss=discriminator_loss):
        super().compile()
        self.generator_optimizer = generator_optimizer
        self.discriminator_optimizer = discriminator_optimizer
        self.generator_loss = generator_loss
        self.discriminator_loss = discriminator_loss

    def get_generator(self):
        return self.generator

    def get_discriminator(self):
        return self.discriminator

    def call(self, inputs, mask=None, training=None):
        if mask is not None:
            inputs = tf.concat([inputs, mask], axis=1)
        generated_samples = self.generator(inputs)
        return generated_samples

    def train_step(self, data):
        # data is a tuple of x_generator and x_discriminator batches
        # drawn from the dataset. The reason behind generating two separate
        # batches of data at each step is that it's how GAIN's algorithm works
        x_gen, x_disc = data

        # =====> Train the discriminator <=====
        for _ in range(self.num_discriminator_steps):
            # Create mask
            mask = tf.constant(1.) - tf.cast(tf.math.is_nan(x_disc), dtype=tf.float32)
            # replace nans with 0.
            x_disc = tf.where(tf.math.is_nan(x_disc), x=0., y=x_disc)
            # Sample noise
            z = self.z_sampler.sample(sample_shape=tf.shape(x_disc))
            # Sample hint vectors
            hint_vectors = self.hint_vectors_sampler.sample(sample_shape=(tf.shape(x_disc)))
            hint_vectors *= mask
            # Combine random vectors with original data
            x_disc = x_disc * mask + (1 - mask) * z
            # Keras model raises `gain`'s weights not created error if we don't implement the
            # `call` method and call it from our code no matter if we compile it or not.
            # so to work around that error we place the call to `Generator`'s `call` method
            # in the `call` method of the `GAIN` model, this way the model's
            # generated_samples = self.generator(tf.concat([x_disc, mask], axis=1))
            generated_samples = self(x_disc, mask=mask)
            # Combine generated samples with original data
            x_hat_disc = (generated_samples * (1 - mask)) + (x_disc * mask)
            with tf.GradientTape() as tape:
                discriminator_pred = self.discriminator(tf.concat([x_hat_disc, hint_vectors], axis=1))
                loss_disc = self.discriminator_loss(discriminator_pred, mask)
            gradients = tape.gradient(loss_disc, self.discriminator.trainable_weights)
            self.discriminator_optimizer.apply_gradients(zip(gradients, self.discriminator.trainable_weights))

        # =====> Train the generator <=====
        mask = tf.constant(1.) - tf.cast(tf.math.is_nan(x_gen), dtype=tf.float32)
        x_gen = tf.where(tf.math.is_nan(x_gen), x=0., y=x_gen)
        z = self.z_sampler.sample(sample_shape=tf.shape(x_gen))
        hint_vectors = self.hint_vectors_sampler.sample(sample_shape=(tf.shape(x_gen)))
        hint_vectors *= mask
        x_gen = x_gen * mask + (1 - mask) * z

        with tf.GradientTape() as tape:
            # generated_samples = self.generator(tf.concat([x_gen, mask], axis=1))
            generated_samples = self(x_gen, mask=mask)
            # Combine generated samples with original/observed data
            x_hat = (generated_samples * (1 - mask)) + (x_gen * mask)
            discriminator_pred = self.discriminator(tf.concat([x_hat, hint_vectors], axis=1))
            loss_gen = self.generator_loss(generated_samples=generated_samples,
                                           real_samples=x_gen,
                                           discriminator_pred=discriminator_pred,
                                           mask=mask,
                                           alpha=self.alpha)
        gradients = tape.gradient(loss_gen, self.generator.trainable_weights)
        self.generator_optimizer.apply_gradients(zip(gradients, self.generator.trainable_weights))

        # Update custom tracking metrics
        self.generator_loss_tracker.update_state(loss_gen)
        self.discriminator_loss_tracker.update_state(loss_disc)

        results = {m.name: m.result() for m in self.metrics}
        results.update({"generator_" + m.name: m.result() for m in self.generator.metrics})
        results.update({"discriminator_" + m.name: m.result() for m in self.discriminator.metrics})
        return results

    def predict_step(self, data):
        """
        Args:
            Transformed data.
        Returns:
            Imputed data that should be reverse transformed
            to its original form.
        """
        if isinstance(data, tuple):
            data = data[0]
        data = tf.cast(data, tf.float32)
        # Create mask
        mask = tf.constant(1.) - tf.cast(tf.math.is_nan(data), dtype=tf.float32)
        data = tf.where(tf.math.is_nan(data), x=0., y=data)
        # Sample noise
        z = self.z_sampler.sample(sample_shape=tf.shape(data))
        x = mask * data + (1 - mask) * z
        # imputed_data = self.generator(tf.concat([x, mask], axis=1))
        imputed_data = self(x, mask=mask)
        imputed_data = mask * data + (1 - mask) * imputed_data
        return imputed_data

    def impute(self,
               x,
               data_transformer=None,
               reverse_transform=True,
               batch_size=None,
               verbose="auto",
               steps=None,
               callbacks=None,
               max_queue_size=10,
               workers=1,
               use_multiprocessing=False
               ):
        """
        Impute function combines GAIN's `predict` method and
        DataTransformer's `reverse_transform` method to fill
        the missing data and transform into the original format.
        It exposes all the arguments taken by the `predict` method.

        Args:
            x: `pd.DataFrame`, dataset with missing values.
            data_transformer: An instance of DataTransformer class which
                was used in transforming the raw input data
            reverse_transform: `bool`, default True, whether to reverse
                transformed the raw imputed data to original format.

        Returns:
            Imputed data in the original format.
        """
        x_imputed = self.predict(x,
                                 batch_size=batch_size,
                                 verbose=verbose,
                                 steps=steps,
                                 callbacks=callbacks,
                                 max_queue_size=max_queue_size,
                                 workers=workers,
                                 use_multiprocessing=use_multiprocessing)
        if reverse_transform:
            if data_transformer is None:
                raise ValueError("To reverse transform the raw imputed data, `data_transformer` must not be None. "
                                 "Please pass the instance of DataTransformer class used to transform the input "
                                 "data as argument to this `impute` method."
                                 "Or alternatively, you can set `reverse_transform` parameter to False, "
                                 "and manually reverse transform the generated raw data to original format "
                                 "using the `reverse_transform` method of DataTransformer instance.")
            x_imputed = data_transformer.reverse_transform(x_imputed)
        return x_imputed

    def get_config(self):
        config = super().get_config()
        new_config = {'generator': keras.layers.serialize(self.generator),
                      'discriminator': keras.layers.serialize(self.discriminator)
                      }
        config.update(new_config)
        return config



# Copyright 2023 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from teras.preprocessing.oenf import PiecewiseLinearEncoding

# CTGAN preprocessing classes
from teras.preprocessing.ctgan import ModeSpecificNormalization
from teras.preprocessing.ctgan import (CTGANDataTransformer,
                                       CTGANDataSampler)

# TVAE preprocessing classes
from teras.preprocessing.tvae import (TVAEDataTransformer,
                                      TVAEDataSampler)


# GAIN preprocessing classes
from teras.preprocessing.gain import (GAINDataTransformer,
                                      GAINDataSampler)


# PCGAIN preprocessing classes
from teras.preprocessing.pcgain import (PCGAINDataTransformer,
                                        PCGAINDataSampler)

# Vime preprocessing classes
from teras.preprocessing.vime import (VimeDataTransformer,
                                      VimeDataSampler)
